"""
File Purpose: solving tfbi
"""
import os

from ...tools import ImportFailed
try:
    import tfbi_theory as tt
except ImportError as err:
    tt = ImportFailed("tfbi_theory", err=err, locals=locals(), abbrv='tt')

from ...defaults import DEFAULTS
from ...errors import InputError, InputMissingError
from ...tools import (
    format_docstring, alias, simple_property,
    UNSET,
    ProgressUpdater,
    xarray_mergeload,
)

_paramdocs_tfbi_solving = {
    'solving_pattern': '''import tfbi_theory as tt
        import PlasmaCalcs as pc
        cc = ... # any PlasmaCalculator object from PlasmaCalcs.
        ds0 = cc.tfbi_ds()
        kp = tt.kPickerLowres(ds0)
        dsk = kp.get_ds()   # copy of ds0, but with ds['k'] = k from kPicker.
        drel = tt.TfbiDisprelC.from_ds(dsk)
        dsR = drel.solve()  # copy of dsk, but with ds['omega'] = solution to TFBI theory!''',
    'solving_pattern_notes': '''if cc has more than ~4 ions, you will want to drop some or group them somehow.
                e.g. for MhdMultifluidCalculator, before calling tfbi_ds():
                    cc.use_mix_heavy_ions(0.3, m_mean_mode='density')
            You can also pick ions directly during cc.tfbi_ds().
                e.g. for Bifrost chromosphere analysis, where n[He_II] < 1e-6 * ne, I use:
                    cc.tfbi_ds(ions=[i for i in cc.fluids if i.q==1 and i!='He_II'])
            After finishing solving, you might want to save the result,
                e.g. dsR.pc.save('filename') saves result to 'filename.pcxarr';
                can load it later via pc.xarray_load('filename.pcxarr').
            If you are solving across "many" points (e.g., more than ~5000),
                you might want to use chunking instead. see TfbiChunkSolver;
                also note cc.tfbi_solver(...) automatically uses TfbiChunkSolver
                if appropriate, i.e. if cc.chunks is set (where cc is a TfbiLoader).''',
    'cc': '''PlasmaCalculator
        PlasmaCalculator object used to load the data.
        Should be a TfbiLoader subclass. (PlasmaCalculator satisfies this by default,
            assuming successful import SymSolver and import tfbi_theory.)''',
    'ions': f'''None or specifier of multiple fluids (e.g. slice, or list of strs)
        None --> use cc.fluids.ions()
        ions are determined when called, not during __init__.
        print warning if this specifies more than DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS ions
            (default: {DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS}), because then solving will be slow and may be inaccurate.''',
    'kres': ''''low', 'mid', or 'high'
        resolution in k-space. Tells which self.kPicker_cls to use.
        'low' --> tfbi_theory.kPickerLowres. Recommended if solving across many (e.g. >1000) points.
        'mid' --> tfbi_theory.kPickerMidres. Recommended if solving across a few (e.g. 10 to 100) points.
        'high' --> tfbi_theory.kPickerHighres. Recommended if solving at only 1 point.''',
    'mod_lmod_ang': '''UNSET or dict
        passed directly to kPicker if provided. Can specify k values other than the defaults.
        see help(self.kPicker_cls) for more details.''',
    'tfbi_all': '''bool
        whether to compute all relevant tfbi vars, ds0 = cc('tfbi_all').
        False --> compute only the necessary vars, ds0 = cc('tfbi_inputs').''',
    'drel_cls': '''None, str, or class
        tfbi_theory class to use for solving TFBI theory.
        None --> use self.drel_cls default: tt.TfbiDisprelC
        str --> use getattr(tt, drel_cls) to get the class.''',
}


### --------------------- TfbiSolver --------------------- ###

@format_docstring(**_paramdocs_tfbi_solving)
class TfbiSolver():
    '''high-level interface for solving TFBI theory across many physical parameters.

    Call TfbiSolver to solve TFBI.
    Example:
        import PlasmaCalcs as pc
        cc = pc.PlasmaCalculator(...)   # <-- your PlasmaCalculator of choice
        solver = pc.TfbiSolver(cc)
        solution = solver()             # alias: solver.solve()
    solution is an xarray.Dataset with all relevant quantities (see TfbiLoader.get_tfbi_all),
        and 'omega' telling roots with largest imaginary part.

    If you need more precise control over the solving process, use the pattern:
        {solving_pattern}

        Notes:
            {solving_pattern_notes}

    TfbiSolver internally stores self.cc, ds0, kp, dsk, drel, and dsR,
        as defined by the pattern above.

    cc: {cc}
    ions: {ions}
    kres: {kres}
    mod, lmod, ang: {mod_lmod_ang}
    tfbi_all: {tfbi_all}
    drel_cls: {drel_cls}
    '''
    kPickerLowres_cls_name = 'kPickerLowres'  # used to make self.kPicker_cls if kres='low'
    kPickerMidres_cls_name = 'kPickerMidres'  # used to make self.kPicker_cls if kres='mid'
    kPickerHighres_cls_name = 'kPickerHighres'  # used to make self.kPicker_cls if kres='high'

    @property
    def kPicker_cls(self):
        '''kPicker class to use for choosing wavevectors to consider. Depends on self.kres:
        'low' --> tfbi_theory.kPickerLowres
        'mid' --> tfbi_theory.kPickerMidres
        'high' --> tfbi_theory.kPickerHighres
        '''
        if self.kres == 'low':
            return getattr(tt, self.kPickerLowres_cls_name)
        elif self.kres == 'mid':
            return getattr(tt, self.kPickerMidres_cls_name)
        elif self.kres == 'high':
            return getattr(tt, self.kPickerHighres_cls_name)
        else:
            raise InputError(f"Invalid kres. Expected 'low' or 'high'; got {self.kres!r}")

    drel_cls_name = 'TfbiDisprelC'  # used to make self.drel_cls.
    # (drel_cls = tt.TfbiDisprelC here would crash immediately if tt import failed.
    # By using drel_cls property, we avoid this crash until it's actually relevant.)
    drel_cls = simple_property('_drel_cls', setdefaultvia='_default_drel_cls',
        doc='''class to use for solving TFBI theory. Default: tt.TfbiDisprelC.''')
    def _default_drel_cls(self):
        '''default drel_cls. getattr(tt, self.drel_cls_name)'''
        return getattr(tt, self.drel_cls_name)

    def __init__(self, cc, ions=None, *, kres='low', mod=UNSET, lmod=UNSET, ang=UNSET, tfbi_all=True,
                 drel_cls=None):
        self.cc = cc
        self.ions = ions
        self.kres = kres
        self.mod = mod
        self.lmod = lmod
        self.ang = ang
        self.tfbi_all = tfbi_all
        if drel_cls is not None:
            if isinstance(drel_cls, str):
                self.drel_cls = getattr(tt, drel_cls)
            elif isinstance(drel_cls, type):
                self.drel_cls = drel_cls
            else:
                raise InputError(f"Invalid drel_cls. Expected None, str, or class; got {drel_cls!r}")

    @property
    def ions_explicit(self):
        '''list of ions from self.cc which would be used during self.cc.tfbi_ds()'''
        if self.ions is None:
            result = self.cc.fluids.ions()
        else:
            result = self.cc.fluids.get(self.ions)
        return result

    def _warn_if_too_many_ions(self):
        '''print warning if self.ions specifies too many ions.'''
        default_nmax = DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS
        ions = self.ions_explicit
        if len(ions) > default_nmax:
            warnmsg = (f">>> Warning: TfbiSolver has {len(ions)} ions, which may be slow and inaccurate. <<<\n"
                       f"    (To avoid this warning, either use fewer ions, or "
                            f"increase DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS (={default_nmax}, currently).)")
            print(warnmsg)

    __call__ = alias('solve')

    @format_docstring(**_paramdocs_tfbi_solving, sub_ntab=1)
    def solve(self, *, verbose=True, **kw_growth_root):
        '''solve TFBI theory. Assigns self.ds0, kp, dsk, drel, and dsR.
        Does not do any caching at this level; if the code crashes you will need to restart it completely.
        For more precise control including possibility for caching, use the pattern directly:
            {solving_pattern}

        verbose: bool
            whether to print progress updates (highly recommended).
        additional kwargs get passed directly to drel.solve().
            (options include: ncpu, ncoarse, careful)
        '''
        updater = ProgressUpdater(print_freq=0 if verbose else -1)  # print all statements if verbose, else nothing.
        self._warn_if_too_many_ions()
        # get ds0 -- manage progress updates here.
        ds0_updater = ProgressUpdater(print_freq=0 if verbose else -1)
        ds0_updater.print(f'getting ds0 via {type(self.cc).__name__}.tfbi_ds() ...')
        self.ds0 = self.cc.tfbi_ds(ions=self.ions, all=self.tfbi_all)
        ds0_updater.finalize(f'solver.ds0 = {type(self.cc).__name__}.tfbi_ds()', end='\n')
        # picking k is usually fast; no need for progress updates.
        kw_kPicker = dict(mod=self.mod, lmod=self.lmod, ang=self.ang)
        kw_kPicker = {k: v for k, v in kw_kPicker.items() if v is not UNSET}
        self.kp = self.kPicker_cls(self.ds0, **kw_kPicker)
        # follow the rest of the pattern. tt manages its own progress updates.
        self.dsk = self.kp.get_ds()
        self.drel = self.drel_cls.from_ds(self.dsk)
        self.dsR = self.drel.solve(verbose=verbose, **kw_growth_root)
        updater.print_clear(force=verbose)  # clear print statements on current line, if we're printing anything.
        updater.finalize(f'{type(self).__name__}.solve() (all steps from start to finish)', always=True)
        return self.dsR

    solution = alias('dsR', doc='''result of self.solve(); alias to self.dsR.''')
    solved = property(lambda self: hasattr(self, 'dsR'), doc='''tells whether self.solution exists''')

    def _repr_contents(self):
        '''return list of strings for contents inside of __repr__'''
        contents = []
        cc_str = f'title={self.cc.title!r}' if hasattr(self.cc, 'title') else '...'
        contents.append(f'cc={type(self.cc).__name__}({cc_str})')
        if self.ions is not None:
            contents.append(f'ions={self.ions!r}')
        if isinstance(self.mod, dict) and len(self.mod) > 0:
            contents.append(f'mod={self.mod!r}')
        if isinstance(self.lmod, dict) and len(self.lmod) > 0:
            contents.append(f'lmod={self.lmod!r}')
        if isinstance(self.ang, dict) and len(self.ang) > 0:
            contents.append(f'ang={self.ang!r}')
        if self.tfbi_all != True:
            contents.append(f'tfbi_all={self.tfbi_all}')
        contents.append(f'solved={self.solved!r}')
        return contents

    def __repr__(self):
        contents = self._repr_contents()
        return f'{type(self).__name__}({", ".join(contents)})'


@format_docstring(**_paramdocs_tfbi_solving, sub_ntab=1)
class TfbiChunkSolver(TfbiSolver):
    '''high-level interface for solving TFBI across many physical parameters, using chunks.
    Chunking details are determined by cc.chunks; see help(cc.chunker) for more info.
    Aside from the chunking, this works very similarly to TfbiSolver.

    Call TfbiChunkSolver to solve TFBI.
    Example:
        import PlasmaCalcs as pc
        cc = pc.PlasmaCalculator(...)   # <-- your PlasmaCalculator of choice
        cc.chunks = dict(...)  # <-- your chunking parameters of choice
        solver = pc.TfbiChunkSolver(cc)
        solution = solver()             # alias: solver.solve()
    solution is an xarray.Dataset with all relevant quantities (see TfbiLoader.get_tfbi_all),
        and 'omega' telling roots with largest imaginary part.

    One good choice of chunks is to subsample spatial dimensions, like:
        XSUBSAMPLE, XBIGSTEP = 64, 256
        XSLICES = [slice(ix0, None, XBIGSTEP) for ix0 in range(0, XBIGSTEP, XSUBSAMPLE)]
        cc.chunks = dict(x_slices=XSLICES)
        # this will do 4 chunks total, each chunk handling every 256th point along x.
        # so, every chunk gives the "full" picture, though very low resolution.
        # but, the chunks are offset from each other by 64, so altogether
        # they actual combine to give every 64th point along x.
        # then, one can rerun later with smaller XSUBSAMPLE (e.g. 32)
        # to improve resolution while re-using existing results!!!

    For more precise control of solving process, consider:
        (copied from TfbiSolver docs): the following pattern & notes: 
            {solving_pattern}

            Notes (copied from TfbiSolver docs):
                {solving_pattern_notes}

        BUT, the pattern above doesn't include chunking.
        The FULL pattern, with chunking included, looks roughly like:
            slicers = cc.chunker().slicers()
            locs = [os.path.join(dst, cc.title_with_slices(s)+'.pcxarr') for s in slicers]
            todo = [(s, loc) for (s, loc) in zip(slicers, locs) if not os.path.exists(loc)]
            for s, p in todo:
                cc.slices = s
                cc.chunks = None
                # (*1) (point of interest, referenced below)
                dsR = ... # <-- copy-paste most of the pattern from above!
                dsR.pc.save(loc)
                # (*2) (point of interest, referenced below)

        [TODO][EFF] would be slightly more efficient to avoid recomputing casc for each chunk.
        However, for "reasonable chunk sizes", this is only a ~10% speedup.

    INPUTS -- used here, not just passed directly to super().
        cc: {cc}
            cc.chunks MUST be nonempty; otherwise, use TfbiSolver instead.
        dst: str
            directory where chunk results should be saved. Internally stored as abspath.
        pre_hook: None or callable of cc
            if provided, called as cc = pre_hook(cc) at point (*1) in the pattern above,
                i.e. before solving each chunk.
            E.g. def pre_hook(cc): cc.tfbi_mask(); return cc
                would apply tfbi_mask() to each chunk before solving.
                (Might want this because slicing after masking is hard,
                but masking after slicing is easy, in present implementation.)
            E.g., to artificially set E_un0_perpmod_B to test constant E/B speed across box:
                def pre_hook(cc):
                    result = cc.copy()
                    EBspeed_grid = pc.xr1d([1000, 3000, 5000, 7000], 'EBspeed')
                    result.set('E_un0_perpmod_B', cc('mod_B') * EBspeed_grid)
                    return result
        post_hook: None or callable of (cc, dsR, loc)
            if provided, called as post_hook(cc, dsR, loc) at point (*2) in the pattern above,
                i.e. after solving each chunk and saving the result.
            E.g., to save a growth rate plot in each chunk's loc directory:
                def post_hook(cc, dsR, loc):
                    dsR.pc.unmask_var('omega').it.growth_kmax().it.growthplot()
                    plt.savefig(os.path.join(loc, 'growthplot'))
                    plt.close()

    INPUTS -- passed directly to super(), i.e. works the same way as TfbiSolver.
        ions: {ions}
        kres: {kres}
        mod, lmod, ang: {mod_lmod_ang}
        tfbi_all: {tfbi_all}
        drel_cls: {drel_cls}
    '''
    def __init__(self, cc, dst, ions=None, *, kres='low', mod=UNSET, lmod=UNSET, ang=UNSET, tfbi_all=True,
                 drel_cls=None, pre_hook=None, post_hook=None):
        if not cc.chunks:
            raise InputMissingError('empty cc.chunks. Set chunks, or use TfbiSolver instead.')
        self.dst = os.path.abspath(dst)
        super().__init__(cc, ions=ions, kres=kres, mod=mod, lmod=lmod, ang=ang,
                         tfbi_all=tfbi_all, drel_cls=drel_cls)
        self.pre_hook = pre_hook
        self.post_hook = post_hook

    def chunk_filepath(self, slicer):
        '''return filepath to use for results from chunk defined by slicer.
        Slicer should be a dict of slices, appropriate as cc.slices = slicer.
        '''
        return os.path.join(self.dst, self.cc.title_with_slices(slicer)+'.pcxarr')

    def slicers_and_filepaths(self):
        '''return list of (slicer, filepath) tuples for all chunks implied by self.cc.chunks.'''
        slicers = self.cc.chunker().slicers()
        filepaths = [self.chunk_filepath(s) for s in slicers]
        return list(zip(slicers, filepaths))

    def done(self):
        '''list of (slicer, filepath) tuples for all chunks which have already been solved.
        (To check if solved, just checks whether filepath exists!)
        '''
        return [(s, p) for (s, p) in self.slicers_and_filepaths() if os.path.exists(p)]

    def todo(self):
        '''list of (slicer, filepath) tuples for all chunks which have NOT yet been solved.
        (To check if solved, just checks whether filepath exists!)
        '''
        return [(s, p) for (s, p) in self.slicers_and_filepaths() if not os.path.exists(p)]

    def n_chunks(self):
        '''total number of chunks implied by self.cc.chunks.'''
        return len(self.cc.chunker())

    @property
    def solved(self):
        '''tells number of chunks solved and total, as "Nsolved_of_Ntotal" string.
        E.g., "3_of_10" if 3 out of 10 chunks have existing solutions.
        '''
        nsolved = len(self.done())
        return f'{nsolved}_of_{self.n_chunks()}'

    def solve(self, *, verbose=2, mergeload=True, **kw_growth_root):
        '''solve TFBI theory using chunking.
        See help(type(self)) for more details.

        Writes results to files, as determined by self.dst.
        Never overwrite any files; assumes pre-existing files are correct solutions.

        verbose: bool or int
            whether to print progress updates (highly recommended).
            0: print nothing
            >=1: print updates about which chunk is being done.
            >=2: print tfbi_solver updates within each chunk.
                (chunks might each take a few minutes)
        additional kwargs get passed directly to drel.solve().
            (options include: ncpu, ncoarse, careful)

        mergeload: bool
            whether to return full solution, mergeloaded across all chunks.
            True --> return full solution.
            False --> return filepath to where chunks are stored (i.e., self.dst).
            Either way, the full solution can later be loaded via pc.xarray_mergeload(dst).
        '''
        updater = ProgressUpdater(print_freq=0 if verbose else -1)
        self._warn_if_too_many_ions()
        # think about chunks:
        todo = self.todo()
        if len(todo) > 0:
            n_done = self.n_chunks() - len(todo)
            updater.print(f'already did {n_done}; need to do {len(todo)}.', end='\n', print_time=False)
        with self.cc.maintaining('chunks', 'slices'):  # restore cc.chunks and cc.slices afterward.
            _chunk_dims = self.cc.chunker().chunk_dims  # (just for prettier print below)
            self.cc.chunks = None  # disable internal chunking within each chunk.
            for i, (slicer, loc) in enumerate(todo):
                _chunk_slices = {d: slicer[d] for d in _chunk_dims if d in slicer}  # (just for pretty print)
                updater.print(f'--- doing {i+1} of {len(todo)}: {_chunk_slices} ---', end='\n')
                self.cc.slices = slicer
                cc = self.cc
                if self.pre_hook is not None:
                    cc = self.pre_hook(self.cc)
                # <-- could just use super(). BUT, might want to customize more later,
                #   e.g. might avoid recomputing casc for each chunk.
                #   (Also, don't want to save ds0, kp, dsk, drel, dsR to self here.)
                #   So, copy the code, instead of calling super().
                ds0_updater = ProgressUpdater(print_freq=0 if verbose>=2 else -1)
                ds0_updater.print(f'getting ds0 via {type(cc).__name__}.tfbi_ds() ...')
                ds0 = cc.tfbi_ds(ions=self.ions, all=self.tfbi_all)
                ds0_updater.finalize(f'solver.ds0 = {type(cc).__name__}.tfbi_ds()', end='\n')
                # picking k is usually fast; no need for progress updates.
                kw_kPicker = dict(mod=self.mod, lmod=self.lmod, ang=self.ang)
                kw_kPicker = {k: v for k, v in kw_kPicker.items() if v is not UNSET}
                kp = self.kPicker_cls(ds0, **kw_kPicker)
                # follow the rest of the pattern. tt manages its own progress updates.
                dsk = kp.get_ds()
                drel = self.drel_cls.from_ds(dsk)
                dsR = drel.solve(verbose=(verbose>=2), **kw_growth_root)
                dsR.pc.save(loc)
                if self.post_hook is not None:
                    self.post_hook(cc, dsR, loc)
        updater.finalize(f'--- {type(self).__name__}.solve() (all chunks) ---',
                         end='\n' if mergeload else '')
        if mergeload:
            return xarray_mergeload(self.dst, verbose=verbose)
        else:
            return self.dst

    def _repr_contents(self):
        '''return list of strings for contents inside of __repr__'''
        contents = super()._repr_contents()
        contents.append(f'dst={self.dst!r}')
        if self.pre_hook is not None:
            contents.append(f'pre_hook={self.pre_hook!r}')
        if self.post_hook is not None:
            contents.append(f'post_hook={self.post_hook!r}')
        return contents



"""
File Purpose: MetadataHaver and UuidHaver
"""
import os
import uuid

from ..errors import FileContentsMissingError
from ..tools import (
    alias, UNSET, ATTR_UNSET,
    datetime_now, read_python_params_file,
)

class MetadataHaver():
    '''class which tracks "metadata" in attrs implied by cls_metadata_attrs.
    E.g., cls_metadata_attrs == {'dirname', 'datetime_run'}.

    To add to cls_metadata_attrs, subclass should use:
        _new_cls_metadata_attrs = {...}

    "metadata" probably refers to "metadata about a simulation run / other data source"
    '''
    cls_metdata_attrs = set()

    def __init_subclass__(cls, **kw_super):
        '''called when subclassing MetadataHaver; sets cls_metadata_attrs,
        from parent's cls_metadata_attrs plus any _new_cls_metadata_attrs provided in cls.
        '''
        super().__init_subclass__(**kw_super)

        inherited = set()
        for parent in cls.__bases__:
            if hasattr(parent, 'cls_metdata_attrs'):
                inherited.update(parent.cls_metdata_attrs)
        new = getattr(cls, '_new_cls_metadata_attrs', set())
        cls.cls_metdata_attrs = inherited.union(new)

    metadata_attrs = alias('cls_metdata_attrs',
        doc='''alias to cls_metadata_attrs; set of all attrs which tell metadata,
        about this data source (simulation run / observation / other data source).
        E.g. metadata_attrs={'dirname', 'datetime_run'}.''')

    @property
    def metadata_all(self):
        '''dict of metadata attributes and their values (value=ATTR_UNSET if missing).
        See also: self.metadata.
        '''
        return {attr: getattr(self, attr, ATTR_UNSET) for attr in self.metadata_attrs}
        # for now, if getattr(self, attr, ATTR_UNSET) crashes, this will crash.
        # eventually, might catch and log crash somewhere instead of killing the whole thing.

    @property
    def metadata(self):
        '''dict of metadata attributes and their values, excluding missing and ATTR_UNSET.
        See also: self.metadata_all.
        '''
        return {attr: v for attr, v in self.metadata_all.items() if v is not ATTR_UNSET}


### --------------------- uuidHaver --------------------- ###

def pc_uuid_here(dir, *, generate=None, default=UNSET, return_meta=False):
    '''return the PlasmaCalcs uuid associated with this directory.

    The uuid is intended to be a universally unique identifier;
        it is highly unlikely that any other uuid will ever match this one.
        This is especially useful for the .register() method and the registry.

    The uuid will be stored in a file named _pc_uuid.txt inside dir, which contains:
        # comment lines starting with '#' explaining the purpose of the file
        # assignment lines below might also have # comments at end of line
        uuid = "uuidstringwithouthyphens"
        generated_at = "YYYY-MM-DDTHH:MM:SS.ssssss+HH:MM"  # when this uuid was generated
        # the file might also contain other assigments but likely will not;
        # other assignments will be ignored.

    dir: str
        path to directory.
    generate: None, True, or False
        whether to generate a new uuid (and create corresponding _pc_uuid.txt file)
        None --> as needed; only if _pc_uuid.txt does not already exist.
        True --> generate new uuid; crash if _pc_uuid.txt already exists.
        False --> return existing uuid; crash (or return default) if _pc_uuid.txt does not exist.
    default: UNSET or any object
        if not UNSET, return `default` if _pc_uuid.txt file does not exist and generate is False.
        Only used if generate is False.
    return_meta: bool
        whether to instead return a dict containing all info from _pc_uuid.txt,
        including the metadata. If True, dict contains 'uuid' and all other keys from file.
        (Values evaluated via ast.literal_eval if possible, else kept as string.)

    When sharing runs across machines, keep the _pc_uuid.txt file,
        to allow PlasmaCalcs to identify it as "the same run" in multiple places.
    When copy-pasting a run as a starting point but wanting PlasmaCalcs
        to treat the new copied run as a different run, delete its _pc_uuid.txt file.
    '''
    file = os.path.join(dir, '_pc_uuid.txt')
    exists = os.path.isfile(file)
    if generate is not None:
        if generate and exists:
            raise FileExistsError(f'generate=True but _pc_uuid.txt file already exists: {file!r}')
        if not generate and not exists:
            if default is not UNSET:
                return default
            raise FileNotFoundError(f'generate=False but _pc_uuid.txt file does not exist: {file!r}')
    if not exists:
        # generate new uuid and write to file
        new_uuid = uuid.uuid4().hex
        generated_at = datetime_now(tz='utc')  # as isoformat string in UTC timezone
        lines = [
            '# PlasmaCalcs-generated identifier for directory contents.',
            '# Delete this file to tell PlasmaCalcs registry to treat this as a new run.',
            f'uuid="{new_uuid}"',
            f'generated_at="{generated_at}"',
        ]
        with open(file, 'w') as f:
            f.write('\n'.join(lines) + '\n')
    # read file, even if just generated it, to ensure consistent behavior.
    params = read_python_params_file(file, eval=True)
    if 'uuid' not in params:
        raise FileContentsMissingError(f"'uuid' missing from _pc_uuid.txt file: {file!r}")
    if return_meta:
        return params
    else:
        return params['uuid']

class UuidHaver(MetadataHaver):
    '''class whose instances have a uuid, gettable via self.get_pc_uuid().
    This is a universally unique identifier for this data source.
    Different instances with the same data source should have the same uuid.

    One simple way to guarantee this behavior is to generate a _pc_uuid.txt file,
    and save it within the data source directory. See pc_uuid_here() for details.
    (However, this class does not assume _pc_uuid.txt exists;
        subclass can choose to implement get_pc_uuid() in a different way if desired.)

    self.pc_uuid should be ATTR_UNSET if it was never generated, else the uuid string.
    (This should be equivalent to self.get_pc_uuid(generate=False, default=ATTR_UNSET)

    To summarize, this class has:
        pc_uuid_here(): get/create uuid for a given directory
        get_pc_uuid(): get/create uuid for this instance's data source. Subclass must implement.
        pc_uuid: uuid if it was ever generated, else ATTR_UNSET.
    '''
    _new_cls_metadata_attrs = {'pc_uuid'}

    pc_uuid_here = staticmethod(pc_uuid_here)

    def get_pc_uuid(self, *, generate=None, default=UNSET):
        '''return the PlasmaCalcs uuid associated with self's data source.

        [Not implemented here; subclass should implement!]

        generate: None, True, or False
            whether to generate a new uuid.
            None --> generate if needed, do not generate if it already exists.
            True --> generate new uuid; crash if it already exists.
            False --> return existing uuid; crash (or return default) if it does not exist.
        default: UNSET or any object
            if not UNSET, return `default` if uuid does not exist and generate is False.
            Only used if generate is False.
        '''
        raise NotImplementedError(f'{type(self).__name__}.get_pc_uuid()')

    @property
    def pc_uuid(self):
        '''PlasmaCalcs uuid (universally unique identifier) associated with self.dirname.
        Returns value of existing uuid if possible, else ATTR_UNSET.

        See also: self.get_pc_uuid(), which will generate the uuid as needed.
        Equivalent to self.get_pc_uuid(generate=False, default=ATTR_UNSET).
        '''
        return self.get_pc_uuid(generate=False, default=ATTR_UNSET)



"""
File Purpose: electric field in Bifrost
"""
import numpy as np
import xarray as xr

from ...errors import FormulaMissingError, LoadingNotImplementedError
from ...quantities import PlasmaDriftsLoader
from ...tools import alias, simple_property


### --------------------- BifrostEfieldLoader --------------------- ###

class BifrostEfieldLoader(PlasmaDriftsLoader):
    '''quantities related to electric field in Bifrost'''

    # # # Q & ETA # # #

    # Q & ETA FOR OHMIC TERM (E_etaJ) #
    @known_var(dims=['snap'])
    def get_qjoule(self):
        '''joule heating (?). eta * |J|**2. Directly from Bifrost aux.
        qjoule in aux is stored in raw units.
            for more aux units see: https://ita-solar.github.io/Bifrost/aux_variables/
        '''
        if 'qjoule' not in self.directly_loadable_vars():
            raise FormulaMissingError('qjoule when not saved to aux')
        ufactor = self.u('energy_density time-1', convert_from='raw')
        return self.load_maindims_var_across_dims('qjoule', u=ufactor, dims=['snap'])

    @known_var(deps=['qjoule', 'mod2_J'])
    def get_eta(self):
        '''eta (scalar), such that E = -u x B + eta J + ....'''
        return self('qjoule') / self('mod2_J')

    # ETA FOR HALL TERM (E_hall) #
    cls_behavior_attrs.register('eta_hall_mode', default='best')
    ETA_HALL_MODE_OPTIONS = {
        'best': '''use 'aux' if possible, else |B|/(|qe| * self('ne', ne_mode='best'))''',
        'aux': '''use 'eta_hall' from aux''',
        'ne_best': '''|B|/(|qe| * self('ne', ne_mode='best'))''',
        'neq': '''|B| / (|qe| * self('ne_neq'))''',
        'table': '''|B| / (|qe| * self('ne_fromtable'))''',
        'ne': '''|B| / (|qe| * self('ne')) using present value of self.ne_mode.''',
    }
    eta_hall_mode = simple_property('_eta_hall_mode', default='best',
        validate_from='ETA_HALL_MODE_OPTIONS',
        doc='''tells where to get eta_hall from. See self.ETA_HALL_MODE_OPTIONS for details.''')

    eta_hall_mode_explicit = alias('eta_hall_mode',
        doc='''self.eta_hall_mode, but if 'best' convert to the explicit mode to use.''')
    @eta_hall_mode_explicit.getter
    def eta_hall_mode_explicit(self):
        mode = self.eta_hall_mode
        if mode == 'best':
            if 'eta_hall' in self.directly_loadable_vars():
                mode = 'aux'
            else:
                mode = 'ne_best'
        return mode

    _ETA_HALL_MODE_EXPLICIT_TO_NE_DEPS = {
        'aux': [],
        'ne_best': ['mod_B', 'abs_qe', ('ne', {'ne_mode': 'best'})],
        'neq': ['mod_B', 'abs_qe', 'ne_neq'],
        'table': ['mod_B', 'abs_qe', 'ne_fromtable'],
        'ne': ['mod_B', 'abs_qe', 'ne'],
    }
    @known_var(dims=['snap'],
               attr_deps=[('eta_hall_mode_explicit', '_ETA_HALL_MODE_EXPLICIT_TO_NE_DEPS')])
    def get_eta_hall(self):
        '''eta_hall (scalar), such that E = -u x B + eta_hall J x Bhat + ....
        eta_hall in aux is stored in raw units.
        eta_hall is equivalent to |B| / (ne * |qe|).

        self.eta_hall_mode controls whether to load from aux or use a formula;
            see self.ETA_HALL_MODE_OPTIONS for options & descriptions.

        result will have 'eta_hall_mode' attr telling the mode used.
            (if 'best', will convert to the explicit mode used, for clarity.)
        '''
        mode = self.eta_hall_mode_explicit
        if mode == 'aux':
            if 'eta_hall' not in self.directly_loadable_vars():
                raise FormulaMissingError('eta_hall when not saved to aux')
            ufactor = self.u('e_field current_density-1', convert_from='raw')
            result = self.load_maindims_var_across_dims('eta_hall', u=ufactor, dims=['snap'])
        else:
            mode_to_ne_mode = {'ne_best': 'best', 'neq': 'neq', 'table': 'table', 'ne': self.ne_mode}
            result = self('(mod_B)/(abs_qe*ne)', ne_mode=mode_to_ne_mode[mode])
        return result.assign_attrs(eta_hall_mode=mode)

    # Q & ETA FOR AMBIPOLAR TERM (E_amb) #
    @known_var(dims=['snap'])
    def get_qjamb(self):
        '''ambipolar joule heating. E_amb dot J.
        qjamb in aux is stored in raw units.
        '''
        if 'qjamb' not in self.directly_loadable_vars():
            raise FormulaMissingError('qjamb when not saved to aux')
        ufactor = self.u('energy_density time-1', convert_from='raw')
        return self.load_maindims_var_across_dims('qjamb', u=ufactor, dims=['snap'])

    cls_behavior_attrs.register('eta_amb_mode', default='best')
    ETA_AMB_MODE_OPTIONS = {
        'best': '''use 'aux' if possible, else 'qJ'.''',
        'aux': '''use 'eta_amb' directly from aux''',
        'qJ': '''compute via J, B, and aux 'qjamb': qjamb = |E_amb dot J|
                --> eta_amb = qjamb / |(((J x Bhat) x Bhat) dot J)|''',
        'qJ2': '''compute estimate via J, B, and aux 'qjamb':
                --> eta_amb = qjamb / |J|**2
                (reasonable if J dot B is small).'''
    }
    eta_amb_mode = simple_property('_eta_amb_mode', default='best',
        validate_from='ETA_AMB_MODE_OPTIONS',
        doc='''tells where to get eta_amb from. See self.ETA_AMB_MODE_OPTIONS for details.''')

    eta_amb_mode_explicit = alias('eta_amb_mode',
        doc='''self.eta_amb_mode, but if 'best' convert to the explicit mode to use.''')
    @eta_amb_mode_explicit.getter
    def eta_amb_mode_explicit(self):
        mode = self.eta_amb_mode
        if mode == 'best':
            if 'eta_amb' in self.directly_loadable_vars():
                mode = 'aux'
            else:
                mode = 'qJ'
        return mode

    _ETA_AMB_MODE_EXPLICIT_TO_NE_DEPS = {
        'aux': [],
        'qJ': ['qjamb', 'J_crosscrossdotself_hat_B'],
        'qJ2': ['qjamb', 'mod2_J'],
    }
    @known_var(dims=['snap'],
               attr_deps=[('eta_amb_mode_explicit', '_ETA_AMB_MODE_EXPLICIT_TO_NE_DEPS')])
    def get_eta_amb(self):
        '''eta_amb (scalar), such that E = -u x B - eta_amb (J x Bhat) x Bhat + ....
        eta_amb in aux is stored in raw units.
        eta_amb physical formula is a bit complicated. See e.g. eqn3 in Martinez-Sykora 2017
            (https://iopscience.iop.org/article/10.3847/1538-4357/aa8866).

        self.eta_amb_mode controls whether to load from aux or use a formula;
            see self.ETA_AMB_MODE_OPTIONS for options & descriptions.

        result will have 'eta_amb_mode' attr telling the mode used.
            (if 'best', will convert to the explicit mode used, for clarity.)
        '''
        mode = self.eta_amb_mode_explicit
        if mode == 'aux':
            if 'eta_amb' not in self.directly_loadable_vars():
                raise FormulaMissingError('eta_amb when not saved to aux')
            ufactor = self.u('e_field current_density-1', convert_from='raw')
            result = self.load_maindims_var_across_dims('eta_amb', u=ufactor, dims=['snap'])
        elif mode == 'qJ':
            result = self('qjamb') / np.abs(self('J_crosscrossdotself_hat_B'))
        elif mode == 'qJ2':
            result = self('qjamb') / self('mod2_J')
        else:
            _known = list(self.ETA_AMB_MODE_OPTIONS.keys())
            raise LoadingNotImplementedError(f'eta_amb_mode: {mode!r}; expected one of {_known}')
        return result.assign_attrs(eta_amb_mode=mode)


    # # # E TERMS # # #

    @known_var(deps=['(SF_u)_cross_B'], aliases=['E_motional'])
    def get_E_uxB(self, *, _B=None):
        '''E_uxB = -u x B, the motional electric field. E = E_uxB + ...
        [EFF] for efficiency, can provide B if already known.
            CAUTION: if providing B, any missing components assumed to be 0.
        '''
        result = -self('(SF_u)_cross_B', _B=_B)
        # don't include fluid=SINGLE_FLUID in result.
        result = result.drop_vars('fluid')  # [TODO] handle 'fluid not in result coords' case?
        return result

    @known_var(deps=['eta', 'J'])
    def get_E_etaJ(self, *, _J=None):
        '''E_etaJ = eta * J. Goes into E = E_uxB + E_etaJ + ...
        [EFF] for efficiency, can provide J if already known.
        '''
        J = self('J') if _J is None else _J
        return self('eta') * J

    @known_var(deps=['eta_hall', 'J_cross_B'])
    def get_E_hall(self, *, _J=None, _B=None, _JxB=None, _mod_B=None):
        '''E_hall = eta_hall * J x B / |B|. Goes into E = E_uxB + E_hall + ...
        [EFF] for efficiency, can provide J, B, JxB, and/or _mod_B if already known.
            CAUTION: if providing vector, any missing components assumed to be 0.
        '''
        eta_hall = self('eta_hall')
        JxB = self('J_cross_B', _J=_J, _B=_B) if _JxB is None else _JxB
        mod_B = self('mod_B', _B=_B) if _mod_B is None else _mod_B
        return eta_hall * JxB / mod_B

    @known_var(deps=['eta_amb', 'J', 'B'])
    def get_E_amb(self, *, _J=None, _B=None, _JxB=None, _mod_B=None):
        '''E_amb = -eta_amb (J x B) x B / |B|**2. Goes into E = E_uxB + E_amb + ...
        [EFF] for efficiency, can provide J, B, JxB, and/or _mod_B if already known.
            CAUTION: if providing vector, any missing components assumed to be 0.
        '''
        eta_amb = self('eta_amb')
        JxB = self('J_cross_B', _J=_J, _B=_B) if _JxB is None else _JxB
        JxBxB = self('(J_cross_B)_cross_B', _val0=JxB, _B=_B)
        mod_B = self('mod_B', _B=_B) if _mod_B is None else _mod_B
        return -eta_amb * JxBxB / mod_B**2


    # # # TOTAL E # # #

    cls_behavior_attrs.register('E_from_aux', default=None)
    E_from_aux = simple_property('_E_from_aux', default=None,
        doc='''whether to load bifrost E from aux (ex ey ez).
        None --> True if ex ey ez all in aux, else False.
        True --> always load from ex ey ez aux (crash if not available)
        False --> never load from ex ey ez aux.''')

    E_from_aux_explicit = alias('E_from_aux',
        doc='''self.E_from_aux, but if None convert to the explicit mode to use.''')
    @E_from_aux_explicit.getter
    def E_from_aux_explicit(self):
        if self.E_from_aux is None:
            return all(x in self.directly_loadable_vars() for x in ['ex', 'ey', 'ez'])
        else:
            return self.E_from_aux

    @known_var(attr_deps=[('E_from_aux_explicit',
                           {True: ['E_uxB'], False: ['E_etaJ', 'E_hall', 'E_amb']})])
    def get_E_u0(self, *, _J=None, _B=None, _JxB=None, _mod_B=None):
        '''E without motional electric field contribution. E_u0 = E_etaJ + E_hall + E_amb.
        if self.E_from_aux_explicit, instead computed via: E_u0 = E (from aux ex ey ez) - E_uxB.

        [EFF] for efficiency, can provide J, B, JxB, and/or _mod_B if already known.
            CAUTION: if providing vector, any missing components assumed to be 0.
        '''
        from_aux = self.E_from_aux_explicit
        if from_aux:
            result = self('E-E_uxB')
        else:
            with self.using(component=None):  # all 3 vector components here
                J = self('J') if _J is None else _J
                B = self('B') if _B is None else _B
                JxB = self('J_cross_B', _J=_J, _B=_B) if _JxB is None else _JxB
                mod_B = self('mod_B', _B=_B) if _mod_B is None else _mod_B
            E_etaJ = self('E_etaJ', _J=J)
            E_hall = self('E_hall', _J=J, _B=B, _JxB=JxB, _mod_B=mod_B)
            E_amb  = self('E_amb',  _J=J, _B=B, _JxB=JxB, _mod_B=mod_B)
            result = E_etaJ + E_hall + E_amb
        return result.assign_attrs(E_from_aux=from_aux)

    @known_var(deps=['E_uxB', 'E_u0'])
    def get_E(self):
        '''electric field from Bifrost. E = E_uxB + E_u0.
        loaded from aux ex ey ez if self.E_from_aux_explicit, else from E_uxB + E_u0.
        '''
        from_aux = self.E_from_aux_explicit
        if from_aux:
            result = self.load_maindims_var_across_dims('e', u='e_field', dims=['snap', 'component'])
        else:
            with self.using(component=None):  # all 3 vector components here
                B = self('B')
            E_uxB = self('E_uxB', _B=B)
            E_u0  = self('E_u0',  _B=B)
            result = E_uxB + E_u0
        return result.assign_attrs(E_from_aux=from_aux)

    E_UN0_MODE_OPTIONS = {**PlasmaDriftsLoader.E_UN0_MODE_OPTIONS,
        'E0_EetaJ_perpB': '''E_un0 = E0_un0_perpB + E_etaJ_perp_B.
            The E0 is E from zeroth-order multifluid physics.
            The E_etaJ is the ohmic term. Physically, ohmic may be included in E0,
            however, in MHD simulations ohmic is primarily from numerical effects,
            and much larger than the physical contribution.''',
        'E0_Eauxu0_perpB': '''E_un0 = E0_un0_perpB + E_u0_perp_B, with E_u0 from aux.
            The E0 is E from zeroth-order multifluid physics.
            The Eaux_u0 is E from aux (ex, ey, ez), minus the motional term, E_uxB.
            This makes some physical sense to do if E from aux includes only E_etaJ,
            which occurs e.g. if Hall & ambipolar physics are disabled,
            in which case the result is conceptually similar to when mode='E0_EetaJ_perpB'.
            Will crash if E not included in aux.''',
    }

    _E_UN0_MODE_TO_TYPE = {**PlasmaDriftsLoader._E_UN0_MODE_TO_TYPE,
        'E0_EetaJ_perpB': 'E0_EetaJ_perpB',
        'E0_Eauxu0_perpB': 'E0_Eauxu0_perpB',
    }

    _E_UN0_TYPE_TO_DEPS = {**PlasmaDriftsLoader._E_UN0_TYPE_TO_DEPS,
        'E0_EetaJ_perpB': ['E0_un0_perpB', 'E_etaJ_perp_B'],
        'E0_Eauxu0_perpB': ['E0_un0_perpB', ('E_u0_perp_B', {'E_from_aux': True})],
    }

    @known_var(value_deps=[('E_un0_type', '_E_UN0_TYPE_TO_DEPS')])
    def get_E_un0(self):
        '''electric field in u_n=0 frame.
        Result depends on self.E_un0_mode; see help(type(self).E_un0_mode) for details.
        Note: using E_un0_mode=None (the default) and self.assume_un='u',
            is equivalent to using E_un0_mode = 'un=u', which will just give self('E_u0').
        Else, E_un0 = E + u_n x B, if can get u_n (crash if can't get u_n).
        '''
        kind = self('E_un0_type').item()
        if kind == 'E+unxB':  # slightly more helpful error than super(); u_neutral will probably crash.
            try:
                un = self('u_neutral', component=None)  # check if can get u_n... probably cant...
            except NotImplementedError:
                errmsg = 'E_un0, when E_un0_mode=None, u_neutral unknown, and self.assume_un != "u".'
                raise FormulaMissingError(errmsg)
            else:
                result = self('E')
                if np.any(un != 0):
                    result = result + self('u_neutral_cross_B', _u_neutral=un)
        elif kind == 'E0_EetaJ_perpB':
            E0_perpB = self('E0_un0_perpB')
            E_etaJ_perpB = self('E_etaJ_perp_B')
            result = E0_perpB + E_etaJ_perpB
        elif kind == 'E0_Eauxu0_perpB':
            E0_perpB = self('E0_un0_perpB')
            E_u0_perpB = self('E_u0_perp_B', E_from_aux=True)
            result = E0_perpB + E_u0_perpB
        else:
            result = super().get_E_un0()
        return result

    @known_var
    def get_E_un0_type(self):
        '''string telling method that will be used to get E_un0. Based on self.E_un0_mode and assume_un.
        possible results include (but may be adjusted further by subclasses):
            'nan' <--> will crash. (e.g. this occurs if E_un0_mode = 'E0_perpmodB_min')
            'E+unxB' <--> self('E') + self('u_neutral_cross_B')
            'E_u0' <--> self('E_u0')
            'E+uxB' <--> self('E') + self('u_cross_B')
            'E' <--> self('E')
            'E0_perpB' <--> self('E0_un0_perpB')
            'E0_EetaJ_perpB' <--> self('E0_un0_perpB') + self('E_etaJ_perpmod_B')
            'E0_Eauxu0_perpB' <--> self('E0_un0_perpB') + self('E_u0_perpmod_B', E_from_aux=True)
        '''
        if self.E_un0_mode is None and self.assume_un == 'u':
            return xr.DataArray('E_u0')
        else:  # super includes lookup in _E_UN0_MODE_TO_TYPE so it includes new types defined above.
            return super().get_E_un0_type()



"""
File Purpose: Component, ComponentList, ComponentDimension, ComponentHaver
"""

from .dimension_tools import (
    DimensionValue, DimensionValueList, string_int_lookup,
    Dimension, DimensionHaver,
)
from ..errors import ComponentKeyError, ComponentValueError, InputError
from ..tools import is_integer, BasisXYZ
from ..tools.xarray_tools import xarray_vectors  # so we can update _BASIS there.

### --------------------- Component & ComponentList --------------------- ###

class Component(DimensionValue):
    '''a single vector component coordinate (e.g. x, y, or z).
    Knows how to be converted to str and int. e.g. 'y' or 1.

    s: string or None
        str(self) --> s. if None, cannot convert to str.
    i: int or None
        int(self) --> i. if None, cannot convert to int.

    Note: not related to "coordinates", e.g. location in space.
    '''
    _dimension_key_error = ComponentKeyError
    _dimension_value_error = ComponentValueError

    def __init__(self, s=None, i=None):
        super().__init__(s, i)


class ComponentList(DimensionValueList):
    '''list of vector component coordinates.'''
    _dimension_key_error = ComponentKeyError
    _dimension_value_error = ComponentValueError
    value_type = Component

    @classmethod
    def from_strings(cls, strings):
        '''return ComponentList from iterable of strings. (i will be determined automatically.)'''
        return cls(Component(s, i) for i, s in enumerate(strings))

    @classmethod
    def from_array(cls, array):
        '''return ComponentList from 0D or 1D array.
        array values should be Component objects, strings, or ints.
            Component objects --> use as-is
            strings --> if 'x', 'y', or 'z', use XYZ.get(string).
                        Else, use Component(s, i=i) for i in range(len(array))
            ints --> if 0, 1, or 2, use XYZ.get(int).
                     Else, use Component(i=i) for i in array
        '''
        # [TODO] use super() instead of repeating code from DimensionValueList.from_array?
        if array.ndim == 0:
            values = [array.item()]
        elif array.ndim == 1:
            values = array.values
        else:
            errmsg = f'{cls.__name__}.from_array expects array.ndim=0 or 1; got ndim={array.ndim}'
            raise DimensionalityError(errmsg)
        result = []
        for i, v in enumerate(values):
            if isinstance(v, cls.value_type):
                result.append(v)
            elif (isinstance(v, str) or isinstance(v, int)) and v in cls.value_type.XYZ:
                result.append(cls.value_type.XYZ.get(v))
            elif isinstance(v, str):
                result.append(cls.value_type(v, i=i))
            elif is_integer(v):
                result.append(cls.value_type(i=v))
            else:
                errmsg = (f'{cls.__name__}.from_array got unexpected value type at index {i}.\n'
                            f'Expected {cls.value_type.__name__}, str, or int; got value={v!r}.')
                raise InputError(errmsg)
        return cls(result)


class ComponentBasisXYZ(BasisXYZ):
    '''simple class to hold X, Y, Z component details.
    (Assumed to be immutable after creation)

    XYZ: ComponentList with length 3
        contains [X, Y, Z] component labels.

    Instances of this class provide attributes:
        X: x component label
        Y: y component label
        Z: z component label
        XHAT: x unit vector, as a DataArray of [1,0,0] with 'component' values XYZ
        YHAT: y unit vector, as a DataArray of [0,1,0] with 'component' values XYZ
        ZHAT: z unit vector, as a DataArray of [0,0,1] with 'component' values XYZ
        YZ_FROM_X: dict telling (y,z) given x, for right-handed coord system x,y,z
            (internally here: string_int_lookup({X:(Y,Z), Y:(Z,X), Z:(X,Y)}))
    '''
    def __init__(self, XYZ):
        if not isinstance(XYZ, ComponentList) or len(XYZ) != 3:
            errmsg = f'ComponentBasisXYZ requires XYZ to be ComponentList of length 3; got {XYZ!r}'
            raise InputError(errmsg)
        super().__init__(XYZ)
        self.YZ_FROM_X = string_int_lookup(self.YZ_FROM_X)
    
    def get(self, x):
        '''return proper component label (i.e. from self.XYZ) given x.
        x can be any valid component specifier, e.g. int, str, Component.
        Equivalent to self.XYZ.get(x).
        '''
        return self.XYZ.get(x)


XYZ = ComponentList.from_strings('xyz')
BASIS_XYZ = ComponentBasisXYZ(XYZ)
_BASIS = BASIS_XYZ

X, Y, Z = _BASIS.XYZ
YZ_FROM_X = _BASIS.YZ_FROM_X

Component.YZ_FROM_X = YZ_FROM_X  # add to Component class for convenience.
Component.XYZ = XYZ  # add to Component class for convenience.

XHAT = _BASIS.XHAT
YHAT = _BASIS.YHAT
ZHAT = _BASIS.ZHAT

# UPDATE tools._BASIS:
xarray_vectors._BASIS = _BASIS


### --------------------- ComponentDimension, ComponentHaver --------------------- ###

class ComponentDimension(Dimension, name='component', plural='components',
                     value_error_type=ComponentValueError, key_error_type=ComponentKeyError):
    '''component dimension, representing current value AND list of all possible values.
    Also has various helpful methods for working with this Dimension.
    '''
    pass  # behavior inherited from Dimension.


@ComponentDimension.setup_haver
class ComponentHaver(DimensionHaver, dimension='component', dim_plural='components'):
    '''class which "has" a ComponentDimension. (ComponentDimension instance will be at self.component_dim)
    self.component stores the current vector component (possibly multiple). If None, use self.components instead.
    self.components stores "all possible vector components" for the ComponentHaver.
    Additionally, has various helpful methods for working with the ComponentDimension,
        e.g. current_n_component, iter_components, take_component.
        See ComponentDimension.setup_haver for details.

    components defaults to XYZ (==ComponentList.from_strings('xyz'))
    '''
    def __init__(self, *, component=None, components=XYZ, **kw):
        super().__init__(**kw)
        if components is not None: self.components = components
        self.component = component



"""
File Purpose: EppicHybridCalculator, for outputs from eppic hybrid simulations.
"""

import xarray as xr

from .eppic_calculator import EppicCalculator
from ...dimensions import IONS
from ...errors import FluidValueError, LoadingNotImplementedError
from ...tools import (
    simple_property,
    xarray_sel,
)

class EppicHybridCalculator(EppicCalculator):
    '''
    EppicHybridCalculator is a subclass of EppicCalculator that is designed to
    handle hybrid simulations in EPPIC. It provides methods for loading and
    processing data specific to hybrid simulations.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hybrid = True

    @known_var(load_across_dims=['fluid'])
    def get_distribution_type(self):
        '''distribution type: DataArray of strings: 'electron', 'ion', or 'neutral'
        (useful internally for handling electrons differently,
            since electrons are fluid while ions are PIC particles.)
        '''
        if self.fluid.is_electron():
            return xr.DataArray('electron')
        elif self.fluid.is_ion():
            return xr.DataArray('ion')
        elif self.fluid.is_neutral():
            return xr.DataArray('neutral')
        else:
            raise FluidValueError(f'expected electron, ion, or neutral fluid; got {self.fluid}')

    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_deltafrac_n(self, *, distribution_type):
        '''normalized density perturbation. deltafrac_n = (n - n0) / n0.
            For hybrid simulations, electron density is the sum of the ion densities  
        '''
        # [TODO][EFF] maybe cache n for ions if recently got it for all ions,
        #    to avoid recalculating, in the "common" case of "get n for all species"...
        if distribution_type == 'electron':
            n0 = self.get_n0()
            n = self('sum_fluids_n', fluid=IONS)
            return (n - n0) / n0
        else:
            return super().get_deltafrac_n()

    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_u(self, *, distribution_type, _ne=None):
        '''velocity. for non-electrons: u = flux / n
        
        For electrons, it comes from the momentum equation for u_e, neglecting du_e/dt:
            The electron momentum equation looks like:
                du_e/dt = -grad(Pe)/(me ne) + (qe/me) (E + u_e x B) - nu_en (u_e - u_n)
            Dropping du_e/dt enables to solve for u_e, using some vector identities.
            Will also ignore u_n (the implementation asserts it is 0).
            Without pressure term, the result would be:
                u_e = (1/(1+Ke^2)) * ((Ke/|B|) * E + (Ke^2/|B|^2) * E x B),
                where Ke = kappae = (qe |B|) / (me nu_en).
            Including the pressure term is "easy", it can be joined with E in the original equation:
                du_e/dt = (qe/me) (E_eff + u_e x B) - nu_en (u_e - u_n), where
                E_eff = E - grad(Pe)/(ne qe).
            Meaning, the full result will be:
                u_e = (1/(1+Ke^2)) * ((Ke/|B|) * E_eff + (Ke^2/|B|^2) * E_eff x B)

        [EFF] optionally, can provide ne as _ne to avoid recalculating it, if already known.
        '''
        if distribution_type == 'electron':
            # assert u_n = 0
            u_n = self('u_n')
            if not (u_n == 0).all():
                raise LoadingNotImplementedError('get_flux for electrons when u_netural != 0')
            # evaluate formula
            n = self('n') if _ne is None else _ne
            # [EFF] calculate E_eff & B first, to avoid recalculating them multiple times.
            #   will need all components, even if len(self.component)==1,
            #   because internally will calculate E AND E cross B
            with self.using(component=None):
                B = self('B')
                E = self('E')
                P = n * self.u('kB') * self('T')
                E_from_P = self.gradient(P) / (n * self('q'))
                E_eff = E - E_from_P
            mod_B = self('mod_B', _B=B)
            skappa = self('skappa')
            E_eff__x = xarray_sel(E_eff, component=self.component)  # E_eff with self.component component(s)
            E_eff_cross_B__x = self('Eeff_cross_B', _Eeff=E_eff, _B=B)
            # ^ could have done self.cross_product(E_eff, B) for full result (all components)
            #   but the way used above enables self to manage components nicely,
            #   e.g. if only 1 component in self.component, result will compute only that component, too.
            u = (1/(1 + skappa**2)) * ((skappa/mod_B) * E_eff__x + (skappa**2/mod_B**2) * E_eff_cross_B__x)
            return u
        else:
            return super().get_u()

    @known_var(partition_across_dim=('fluid', 'distribution_type'))  # [TODO] deps (for electrons)
    def get_flux(self, *, distribution_type):
        '''flux. (for non-electrons: directly from EPPIC. for electrons: from momentum equation)

        For electrons, get u from momentum equation for u_e, neglecting du_e/dt.
        See self.help('u') (or help(self.get_u)) for more details.
        Then, flux_e = n_e * u_e
        '''
        if distribution_type == 'electron':
            n = self('n')
            u = self('u', _ne=n)
            return n * u
        else:
            return super().get_flux()
    
    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_n0(self, *, distribution_type):
        '''background density. (directly from Eppic.)
        For hybrid simulations, electron density is the sum of the ion densities  
        note: as with all other quantities, this will be output in [self.units] unit system;
            numerically equal to eppic.i value if using 'raw' units.
        '''
        if distribution_type == 'electron':
            return self('sum_fluids_n0', fluid=IONS)
        else:
            return super().get_n0()

    cls_behavior_attrs.register('_Te_fix_factor', default=2/3)
    _Te_fix_factor = simple_property('_Te_fix_factor_', default=2/3,
        doc='''factor to multiply electron temperature by, to correct for issue in hybrid EPPIC.
        (if hybrid EPPIC "temperature" output gets fixed, can set this to 1,
        and maybe pick a different default / fancier behavior to try to distinguish whether
            you are working with a pre-fix output or post-fix output,...
            maybe based on file date for the snapshot files? Or, some other flag in eppic.i...)''')
        
    @known_var(partition_across_dim=('fluid', 'distribution_type'), aliases=['temp'])
    def get_T(self, *, distribution_type):
        ''' temperature in Kelvin.

        For electrons, just loads 'temperature' from snapshot.
            NOTE: also, multiplies by self._Te_fix_factor, default 2/3,
                to account for issue with "temperature" outputs in hybrid EPPIC.
        For ions, equivalent to rmscomps_Ta.
            (Use self.T_indim_only=True if you want to ignore Ta_z in 2D sim)
        '''
        # [TODO] actually, should probably define get_Ta_or_Tajoule for electrons,
        #    since that is the "base" var in super().
        #    e.g. I think Ta for electrons could just return 'temperature' (without varying across component);
        #    that might make it just work, but it might also take some debugging to get it right.
        #    - SE (2025/05/07)
        if distribution_type == 'electron':
            result = self.load_maindims_var_across_dims('temperature', dims=['snap']) / self.u('kB')
            result = result * self._Te_fix_factor
            return result
        else:
            return super().get_T()
    
    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_T_box(self, *, distribution_type):  # [TODO] deps (for electrons)
        '''temperature of the entire simulation box, as if full of particles,
        and observed by something that could not resolve the individual cells.
        Equivalent: rmscomps(Ta_from_moment2).

        Ignores Ta components for directions in which the box has no extent.

        NOTE: electron temperatures are handled carefully, via the description above.
            Since we only know the isotropic T for electrons in each cell,
            the formula to get T_box for electrons is a bit more complicated:

            T_x = nmean_T + m * (nmean_(u_x^2) - nmean_(u_x)^2)
            T_box = rmscomps(T), e.g.:
                if 3D in x,y,z: sqrt(T_x^2 + T_y^2 + T_z^2) / sqrt(3)
                if 2D in x,y:   sqrt(T_x^2 + T_y^2) / sqrt(2)
            where nmean is the density-weighted mean over the entire box,
            i.e. mean(n * quantity) / mean(n).

            see Evans+2025 Appendix C for more details, including the derivation.
        '''
        if distribution_type == 'electron':
            ## naive implementation:
            # return self('mean_T')  # Te is isotropic for hybrid eppic --> T_box is just mean_T.

            ## careful implementation (see Evans+2025 Appendix C)
            with self.using(component=self.maindims):   # e.g., ignore u_z if 2D in x,y
                # [TODO][EFF] respect self.stats_dimpoint_wise for nmean computations.
                #   (useful if arrays are very large.)
                T = self('T')
                m = self('m')
                n = self('n')
                u = self('u', _ne=n)
                n0 = n.mean(dim=self.maindims)
                nmean_T = (n * T).mean(dim=self.maindims) / n0
                nmean_ux2 = (n * u**2).mean(dim=self.maindims) / n0
                nmean_ux = (n * u).mean(dim=self.maindims) / n0
                Tx = nmean_T + m * (nmean_ux2 - nmean_ux**2)
                return self.rmscomps(Tx)
        else:
            return super().get_T_box()



"""
File Purpose: basic tests for bifrost hookup in PlasmaCalcs.
These tests require some simulation output.
"""
import os

import numpy as np

import PlasmaCalcs as pc

HERE = os.path.dirname(__file__)


### --------------------- can create bifrost calculator --------------------- ###

def get_bifrost_calculator(**kw_init):
    '''make a bifrost calculator'''
    with pc.InDir(os.path.join(HERE, 'test_bifrost_tinybox')):
        bc = pc.BifrostCalculator(**kw_init)
    return bc

def get_bifrost_multifluid_calculator(**kw_init):
    '''make a multifluid bifrost calculator (for multifluid analysis of bifrost outputs)'''
    with pc.InDir(os.path.join(HERE, 'test_bifrost_tinybox')):
        bm = pc.BifrostMultifluidCalculator(**kw_init)
    return bm

def test_make_bifrost_calculator():
    '''ensure can make bifrost calculator and bifrost multifluid calculator'''
    get_bifrost_calculator()
    get_bifrost_multifluid_calculator()


### --------------------- misc tests --------------------- ###

def test_bifrost_calculator_misc():
    '''load some values. Mostly trying to ensure things don't crash.'''
    bc = get_bifrost_calculator()
    ## check: can load all simple MHD vars:
    bc('e')
    bc('r')
    bc('B')
    ## check: can load derived vars:
    assert bc.eos_mode == 'neq'
    bc('T')  # <-- comes directly from aux outputs.
    bc('P')  # <-- plugs e & r into lookup tables.
    bc.tabin['P']  # <-- ensure P table actually exists...
    ## check: misc. dimensionality things:
    assert bc('B').sizes['component'] == 3
    assert set(bc.maindims) == set(['x', 'z'])  # this is a 2D sim without y extent.
    assert bc('r').sizes['x'] == bc.maindims_sizes['x'] == 50
    assert bc('r').sizes['z'] == bc.maindims_sizes['z'] == 50
    ## check: misc. things related to "single fluid"
    assert bc('r').equals(bc('SF_r'))
    assert np.allclose(bc('m') * bc('n') / bc('r'), 1)

def test_bifrost_multifluid_calculator_misc():
    '''load some values. Mostly trying to ensure things don't crash.'''
    bm = get_bifrost_multifluid_calculator()
    assert len(bm.elements) == 16
    assert len(bm.fluids) == 19  # 16 once-ionized, plus e, H_I, and He_III.
    n = bm('n')
    assert n.sizes['fluid'] == 19
    assert bm('SF_n')['fluid'].item() is pc.SINGLE_FLUID
    # nusn = bm('nusn')  # TODO: currently, crashes; see issue #21.
    bm.snap  = 0  # using non-iterable bm.snap is the workaround for now.
    nusn = bm('nusn')
    assert nusn.sizes['fluid'] == 19
    assert nusn['jfluid'] == 'H_I'
    assert np.all(nusn.pc.sel(fluid='H_I') == 0)  # 0 self-self collisions
    assert nusn.max() > 1e5  # sometimes collisions are significant;
    assert 1e-4 < nusn.pc.sel(fluid=pc.CHARGED).min() < 1e-2  # sometimes they aren't.
    # [TODO] more misc. checks...


### --------------------- stagger tests --------------------- ###

def test_bifrost_stagger():
    '''test some things related to stagger operations'''
    bc = get_bifrost_calculator()
    ## check: stagger mesh interpolation operations for B.
    staggered = bc('B')
    raw = bc('B', stagger_direct=False)
    # (By (centered) = shift(By, yup), but can't do yup for 2D without y)
    assert staggered.pc.sel(component='y').equals(raw.pc.sel(component='y'))
    # (Bx (centered) = shift(Bx, xup), which has some impact on results.
    #  similar for Bz. Tested impact on results once when code is working for sure,
    #  and hard-coded answers below. If deviation changes significantly,
    #  due to changes in stagger algorithm itself, this will crash.
    #  putting a small margin of error here in case of tiny numerical rounding errors.)
    ratx = staggered.pc.sel(component='x') / raw.pc.sel(component='x')
    ratz = staggered.pc.sel(component='z') / raw.pc.sel(component='z')
    assert 0.005 < ratx.std()  < 0.020
    assert 0.900 < ratx.min()  < 0.930
    assert 0.990 < ratx.mean() < 1.010
    assert 1.000 < ratx.max()  < 1.030
    assert 0.0005 < ratz.std()  < 0.0009
    assert 0.9950 < ratz.min()  < 0.9970
    assert 1.0005 < ratz.mean() < 1.0020
    assert 1.0060 < ratz.max()  < 1.0090
    ## check: stagger mesh interpolation and derivative operations for J.
    # bc('J')  # TODO: currently, crashes; see issue #21.
    bc.snap = 0  # using non-iterable bc.snap is the workaround for now.
    J = bc('J')
    rawJ = bc('J', J_stagger=False)
    Jdiff = J - rawJ
    # (similarly to above: stagger has some impact on results.
    #  hard-coded; test will crash if updates ever significantly change results.
    #  using values directly also helps to check that units are correct.)
    assert 0.002 < Jdiff.std() < 0.003
    assert -0.007 < Jdiff.min() < -0.006
    assert 0.000 < Jdiff.mean() < 0.001
    assert 0.025 < Jdiff.max() < 0.031
    ## check: using stagger interface separately.
    bc = get_bifrost_calculator()
    # (stagger interface expects: raw units and 3D numpy arrays.)
    bc.snap = 0
    bc.units = 'raw'
    bc.squeeze_direct = False
    J_z_from_interface = bc.stagger(bc('B_y').values, 'ddxdn xup yup')
    assert np.allclose(1, J_z_from_interface / bc('J_z'))
    ## check: bc.stagger_minimal_slicing = True doesn't cause crash.
    # (TODO: check values for correctness as well...)
    bc = get_bifrost_calculator(snap=0)
    bc.stagger_minimal_slicing = True
    bc('B')
    bc('J')
    # check stagger minimal slicing gives same answers as non-minimal slicing
    bc.slices = dict(x=slice(10,30))  # note, can't start at 0 or end at -1 yet.
    Borig = bc('B', stagger_minimal_slicing=False)
    Bstag = bc('B', stagger_minimal_slicing=True)
    assert Borig.equals(Bstag)
    bc.slices = dict(x=15)  # note, need step >= 10
    Jorig = bc('J', stagger_minimal_slicing=False)
    Jstag = bc('J', stagger_minimal_slicing=True)
    assert Jorig.equals(Jstag)
    # [TODO] why does it get messed up when using slice with step???
    #   the results aren't the same; sometimes rat1 (below) will be >10% even within this tiny box...
    # doing with step gives not-exactly-the-same results, but results are extremely close.
    #  (not sure why not exactly the same. But "extremely close" should be good enough.
    #   seems like floating rounding differences, maybe from doing ops in different order?)
    # bc.slices = dict(z=slice(10, 42, 15))  # note, need step >= 10
    # Jorig = bc('J', stagger_minimal_slicing=False)
    # Jstag = bc('J', stagger_minimal_slicing=True)
    # Jnaive = bc('J', J_stagger=False)
    # rat0 = np.abs((Jorig - Jnaive)/Jorig)
    # assert rat0.median().item() > 0.5  # often more than 50% difference between no-stagger and stagger J.
    # rat1 = np.abs((Jorig - Jstag)/Jorig)
    # assert rat1.median().item() < 0.001  # with/without minimal slicing, differs by less than 0.1% usually.
    # assert rat1.max().item() < 0.15  # with/without minimal slicing, differes by less than 15% always.

    # ANSWER ^^ it has nothing to do with step! it is due to flip_z_mesh;
    #  applying flip_z_mesh but not also changing signs means ddz_(quant) is negative.
    #  see issue $42 for more details.

# [TODO] more bifrost tests...



"""
Package Purpose: misc. science tools
Like all tools, these should be portable, i.e. not specific to PlasmaCalcs.
These generally solve a very particular science problem (like ray draping),
    rather than doing more generic science-related computations (like curve fitting).
"""
from .ray_draping import (
    RayDrapeScene, RayDrapeAnchor, RayDrapeFOV, RayDrapingCore, RayDraping,
)



"""
File Purpose: reading runtime info for EppicCalculator

this is the timing information from the timers.dat file.
"""
import numpy as np
import xarray as xr

from .eppic_io_tools import (
    read_timers_dat, eppic_clock_times_here,
    eppic_earliest_start_time_from_jobfiles,
)
from ...errors import FileContentsError
from ...quantities import QuantityLoader
from ...tools import (
    alias, ATTR_UNSET,
    find_jobfiles,
    xarray_promote_dim, xarray_sum,
    datetime_str,
)

class EppicRuntimeInfoLoader(QuantityLoader):
    '''runtime info for EppicCalculator.'''


    # # # SIMULATION START TIME # # #

    @property
    def datetime_run(self):
        '''datetime corresponding to roughly when this simulation run was started.
        As isoformat string like: "YYYY-MM-DDTHH:MM:SS.ssssss+00:00" (UTC timezone).
        CAUTION: this is a best-effort estimate which may vary across operating systems.

        Result is earliest start time found amongst all jobfiles in self.dirname if any exist,
        else file system timestamp of first snapshot in self.snapdir if any exist,
        else ATTR_UNSET.
        '''
        try:
            result = eppic_earliest_start_time_from_jobfiles(self.dirname)
        except (FileNotFoundError, FileContentsError):
            result = ATTR_UNSET
        else:
            result = datetime_str(result, sep='T', tz='utc')
        if result is ATTR_UNSET:  # no jobfile with time.
            result = super().datetime_run
        return result


    # # # FROM LOG FILE # # #

    def jobfiles(self):
        '''return list of all jobfiles within self.dirname directory.'''
        return find_jobfiles(self.dirname)

    logfiles = alias('jobfiles')

    def clock_times(self):
        '''return dict of clock times from this run. (From reading self.jobfiles.)
        Result can have keys:
            'start': datetime telling when the run started.
            'stepstart': datetime telling when the iterations started.
            'end': datetime telling when the run ended.
            'init_seconds': (stepstart - start) [seconds]
            'steps_seconds': (end - stepstart) [seconds]
            'total_seconds': (end - start) [seconds]
        If jobfile missing any info, relevant keys will not appear in result.
        '''
        return eppic_clock_times_here(self.dirname)

    @known_var
    def get_clock_times(self):
        '''dataset of clock times from this run. (From reading self.jobfiles.)
        Result can have keys:
            'start': datetime telling when the run started.
            'stepstart': datetime telling when the iterations started.
            'end': datetime telling when the run ended.
            'init_seconds': (stepstart - start) [seconds]
            'steps_seconds': (end - stepstart) [seconds]
            'total_seconds': (end - start) [seconds]
        If jobfile missing any info, relevant keys will not appear in result.
        '''
        return xr.Dataset(self.clock_times())

    @known_var(deps=['clock_times'])
    def get_init_seconds(self):
        '''time [in seconds] spent initializing the run. (between start and when steps start.)'''
        return self('clock_times')['init_seconds']

    @known_var(deps=['clock_times'])
    def get_steps_seconds(self):
        '''time [in seconds] spent doing timesteps. (total duration minus init_seconds.)'''
        return self('clock_times')['steps_seconds']

    @known_var(deps=['clock_times'], aliases=['total_time_seconds'])
    def get_total_seconds(self):
        '''duration [in seconds] spent to run the entire run.'''
        return self('clock_times')['total_seconds']


    # # # FROM TIMERS.DAT # # #

    def timers_dat(self, *, with_snaps=False, as_array=False):
        '''return timers.dat as an xarray.Dataset. (dimension will be named 'it')
        result will have the same units as timers.dat file.

        with_snaps: bool
            if True, attach snap & t coords and promote 'snap' to main dim.
            based on self.snaps (not self.snap)
        as_array: bool
            whether to use xarray.Dataset.to_array() to return a DataArray instead of Dataset.
            if True, vars from Dataset will be concatenated along the new dimension named 'timer'.
        '''
        result = read_timers_dat(self.dirname, as_array=as_array, # [TODO][EFF] use caching if this is slow.
                                fix_snaps=(self.snaps_from=='parallel'))
        if with_snaps:
            existing_snaps = self.existing_snaps()
            result = self.assign_snap_along(result, 'it', existing_snaps)
            result = xarray_promote_dim(result, 'snap')
            if len(existing_snaps) != len(self.snaps):  # some self.snaps refer to MISSING_SNAP
                # fill MISSING_SNAP points with NaN.
                result = result.reindex(snap=self.snaps, fill_value=self.snap_dim.NAN)
                # assign t coords to MISSING_SNAP points  # [TODO][EFF] do something more efficient?
                result = self.assign_snap_along(result, 'snap', self.snaps)
            result = result.drop_vars('it')
        return result

    @known_var(dims=['snap'])
    def get_timers(self):
        '''timers_dat info as an xarray.Dataset, at snaps in self.snap. see also: 'runtimes'.'''
        result = self.timers_dat(with_snaps=True, as_array=False)
        result = result.sel(snap=np.array(self.snap))
        return result

    @known_var(dims=['snap'])
    def get_runtimes(self):
        '''timers_dat info as an xarray.DataArray, at snaps in self.snap. see also: 'timers'.'''
        result = self.timers_dat(with_snaps=True, as_array=True)
        result = result.sel(snap=np.array(self.snap))
        return result

    @known_var(deps=['timers'])
    def get_run_time(self):
        '''Wall clock runtime for each snap. Same units as timers_dat info.'''
        return self('timers')['Wall Clock']

    @known_var(deps=['timers'])
    def get_write_time(self):
        '''time spent writing output files. Same units as timers_dat info.'''
        return self('timers')['output']

    @known_var(deps=['run_time', 'write_time'])
    def get_calc_time(self):
        '''Wall clock runtime, ignoring time spent writing output files. Same units as timers_dat info.'''
        return self('run_time') - self('write_time')

    @known_pattern(r'(write|calc)_time_frac', deps=['run_time', {0: '{group0}_time'}])
    def get_time_frac(self, var, *, _match=None):
        '''fraction of runtime spent on writing or calculating.
            var='write_time_frac' --> fraction of runtime spent writing output files.
            var='calc_time_frac' --> fraction of runtime spent calculating, ignoring write_time.
        '''
        here, = _match.groups()
        return self(f'{here}_time') / self('run_time')

    @known_pattern(r'(run|calc|write)_(timestep|dt)_cost(_f|_nosub|_fnosub|_nosubf)?',
            deps=['nit_since_prev', 'npd', {0: '{group0}_time'},
                {2: lambda groups: 'subcycle' if (groups[2] is not None and 'sub' in groups[2]) else []}])
    def get_timestep_cost_or_dt_cost(self, var, *, _match=None):
        '''total cpu time per simulated particle, per timestep (or per dt).
        
        time_cost = (runtime / timestep_or_dt) * (n_processors / total number of particles)
        total number of particles = n_processors * npart.
            Note: n_processors cancels out; time_cost = (runtime / timestep_or_dt) / npart
        timestep_or_dt = one timestep or one dt; see below.
        npart = number of simulated particles, in one processor. Depends on {settings}; see below.

        '{clock}_{time}_cost{settings}'
            E.g. 'run_timestep_cost', 'write_dt_cost_f', 'calc_dt_cost_nosubf'
            {clock} = 'run', 'calc', or 'write'
                tells which clock to use.
                'run' --> 'Wall clock' | 'calc' --> 'Wall Clock - output' | 'write' --> 'output'
            {time} = 'timestep' or 'dt'
                'timestep' --> report result as cost per timestep, regardless of dt.
                'dt'       --> report result as cost per dt (converted to SI units).
            {settings} = '', '_f', '_nosub', '_fnosub', or '_nosubf'
                tells whether to return a separate value for each fluid, and whether to account for subcycling.
                    ''        --> single value. account for subcycling.     npart = self('npd/subcycle').sum('fluid')
                    '_nosub'  --> single value. ignore subcycling.          npart = self('npd').sum('fluid')
                    '_f'      --> per-fluid values. account for subcycling. npart = self('npd/subcycle')
                    '_fnosub' --> per-fluid values. ignore subcycling.      npart = self('npd')
                    '_nosubf' --> same as '_fnosub'; provided for convenience.

                accounting for subcycling means dividing by the subcycling factor,
                because less effort is spent on subcycled distributions.
        '''
        clock, time, settings = _match.groups()
        clock_per_timestep = self(f'{clock}_time') / self('nit_since_prev')
        # get npart
        if settings is None:
            settings = ''
        if 'f' in settings:
            npart = self('npd')
            if 'nosub' not in settings:
                npart = npart / self('subcycle')
        else:
            with self.using(fluid=None):
                npart = self('npd')
                if 'nosub' not in settings:
                    npart = npart / self('subcycle')
                npart = xarray_sum(npart, 'fluid')
        result = clock_per_timestep / npart
        if time == 'dt':
            result = self.record_units(result / self.input_deck['dt'])
        elif time == 'timestep':
            pass  # already divided by timestep.
        else:
            raise NotImplementedError(f'coding error, expected time="timestep" or "dt", got {time!r}')
        return result


    # # # CONVERTING BETWEEN TIMERS.DAT <--> SECONDS # # #

    @known_var(deps=['run_time', 'steps_seconds'])
    def get_timer2seconds(self):
        '''conversion factor from timers.dat units to seconds.
        E.g. timer2seconds = 0.01 <--> 1 timer unit = 0.01 seconds <--> 1 second = 100 timer units.

        timer2seconds = steps_seconds / sum(run_time)
        Might vary from machine to machine. Seems to be 0.01 on Frontera (plus small rounding errors).
        '''
        return self('steps_seconds') / self('run_time').sum()

    @known_var(deps=['timer2seconds'])
    def get_seconds2timer(self):
        '''conversion factor from seconds to timers.dat units.
        E.g. seconds2timer = 100 <--> 1 timer unit = 0.01 seconds <--> 1 second = 100 timer units.

        seconds2timer = sum(run_time) / steps_seconds
        Might vary from machine to machine. Seems to be 100 on Frontera (plus small rounding errors).
        '''
        return 1 / self('timer2seconds')


    # # # TRYING TO "GUESS" RUNTIME PER TIMESTEP # # #

    # stats from some (~100) 2D runs I (sam) have on my machine,
    #   when computing 'log10_meant_cpu_seconds_per_each'.
    #   Excluded runs with ncells_sim <= 128**2 (because small runs timing isn't as important to get right)
    #       and runs with nx >= 20 (because I don't have enough runs to learn dependence with nsubdomains < max,
    #           and max nsubdomains is when nx = 8. Closest-to-optimal safe nsubdomains is when nx = 16.)
    CPU_SECONDS_PER_EACH_INFO = {
        'timers': ['vadv time', 'xadv time', 'charge', 'collect', 'efield', 'output'],
        'timer_shorthand': ['vadv', 'xadv', 'charge', 'collect', 'efield', 'output'],
        'stats': {
            'min': [-7.60274492, -8.08801379, -8.29066597, -6.4194993 , -6.45974422, -9.13447354],
            'mean': [-7.14767658, -7.71036947, -8.16571132, -5.41090113, -5.84219558, -8.26272736],
            'median': [-7.14165344, -7.7622515 , -8.18222028, -5.4868952 , -5.83410388, -8.29974968],
            'max': [-6.64987625, -6.85562141, -7.85642911, -4.26655371, -5.16467001, -6.96257236],
            'std': [0.1618933 , 0.24986124, 0.07723055, 0.39052264, 0.27228036, 0.41827969],
        },
        # arr.polyfit(log10_n_nodes, degree=1) where arr==log10_meant_cpu_seconds_per_each.
        'polyfit_n_nodes': {
            'coeff': {   # coeffs for each degree. Direct result of polyfit.
                1: [ 0.07367062,  0.33485588,  0.07118243,  0.98491967,  0.76134439,  0.42652909],
                0: [-7.23738158, -8.11810654, -8.25238656, -6.61018799, -6.76924614, -8.78209026],
            },
            'std': {   # stddev on coeffs for each degree. Larger stddev --> less certainty on coeff result.
                1: [0.05228419, 0.07328888, 0.02404283, 0.07154953, 0.03530289, 0.12873761],
                0: [0.06593647, 0.09242584, 0.03032081, 0.09023231, 0.04452106, 0.16235315],
            },
        },
        # which timers we should actually use polyfit_n_nodes for, when guessing runtime.
        #   default is ['collect', 'efield'] because those have smaller std & (max-min) after detrending
        #   (and since the fit is in logspace, that should directly cause smaller error in predictions).
        'polyfit_n_nodes_apply': ['collect', 'efield'],
    }

    @known_pattern(r'guess_cpu_seconds_(all|vadv|xadv|charge|collect|efield|output)_per_each(_safe|)',
                   deps=['n_nodes'])
    def get_guess_cpu_seconds_per_each(self, var, *, _match=None):
        '''guess log10(cpu seconds per (particle, per) grid cell, per timestep), for the indicated timer.
        if 'safe' included in var, uses mean + stddev instead of just mean, to get a conservative estimate.
        Result excludes initialization time (before iterations begin).

        timer options:
            'all': result will be a Dataset of results from all timers, with timer names as data_vars.
                    (timer names are 'vadv time', 'xadv time', 'charge', 'collect', 'efield', 'output'.)
            'vadv', 'xadv', 'charge', 'collect', 'efield', 'output': result for corresponding timer.

        if timer is 'all', return sum of the results from other timers.

        These results are per particle per grid cell per timestep:
            'vadv time', 'xadv time', 'charge', 'output'.
        These results are per particle per timestep:
            'collect', 'efield'.

        The guess is informed by self.CPU_SECONDS_PER_EACH_INFO,
            which contains numerical values derived from some simulations run on Frontera.
        Assumes 'output' is not very expensive.
            The "output" timer was not tested across a large variety of 'nout'.
            (trying to detrend with respect to nout didn't affect stddev significantly.)
        '''
        here, safe = _match.groups()
        INFO = self.CPU_SECONDS_PER_EACH_INFO
        if here == 'all':
            result = {}
            for i, timer_shorthand in enumerate(INFO['timer_shorthand']):
                timer = INFO['timers'][i]
                result[timer] = self(f'guess_cpu_seconds_{timer_shorthand}_per_each{safe}')
            return xr.Dataset(result)
        # else, getting result for an individual timer.
        i = INFO['timer_shorthand'].index(here)
        timer = INFO['timers'][i]
        if timer in INFO['polyfit_n_nodes_apply']:
            log_n_nodes = self('log10_n_nodes')
            pfit = INFO['polyfit_n_nodes']
            coeff1 = pfit['coeff'][1][i]
            coeff0 = pfit['coeff'][0][i]
            if safe:
                coeff1 = coeff1 + pfit['std'][1][i]
                coeff0 = coeff0 + pfit['std'][0][i]
            guess = coeff1 * log_n_nodes + coeff0
        else:
            guess = INFO['stats']['mean'][i]
            if safe:
                guess = guess + INFO['stats']['std'][i]
        # above, got log10(guess). Exponentiate and return:
        return xr.DataArray(10**guess)

    @known_pattern(r'guess_cpu_seconds_(all|sum|vadv|xadv|charge|collect|efield|output)_per_(timestep|ct)(_safe|)',
                   deps=[{1: 'cpu_per_each2{group1}'}, {2: 'guess_cpu_seconds_all_per_each{group2}'}])
    def get_guess_cpu_seconds_per_timestep_or_ct(self, var, *, _match=None):
        '''guess cpu seconds required per (cell, per) timestep, for the indicated timer.
        if 'safe' included in var, uses mean + stddev instead of just mean, to get a conservative estimate.
        Use 'ct' to indicate 'per cell per timestep'.
            (result using 'timestep') == ncells_sim * (result using 'ct').
        Result excludes initialization time (before iterations begin).

        timer options:
            'all': result will be a Dataset of results from all timers, with timer names as data_vars.
                    (timer names are 'vadv time', 'xadv time', 'charge', 'collect', 'efield', 'output'.)
            'sum': result will be the sum of the results from other timers.
            'vadv', 'xadv', 'charge', 'collect', 'efield', 'output': result for corresponding timer.

        Assumes 'output' is not very expensive; it was not tested as much as the others.
        The guess is informed by self.CPU_SECONDS_PER_EACH_INFO.
        For more details see: guess_cpu_seconds_per_each.
        '''
        here, ct, safe = _match.groups()
        INFO = self.CPU_SECONDS_PER_EACH_INFO
        if here == 'all' or here == 'sum':
            result = {}
            for i, timer_shorthand in enumerate(INFO['timer_shorthand']):
                timer = INFO['timers'][i]
                result[timer] = self(f'guess_cpu_seconds_{timer_shorthand}_per_{ct}{safe}')
            if here == 'all':
                return xr.Dataset(result)
            else:  # here == 'sum'
                return sum(result.values())
        # else, getting result for an individual timer.
        i = INFO['timer_shorthand'].index(here)
        timer = INFO['timers'][i]
        guess_each = self(f'guess_cpu_seconds_{here}_per_each{safe}')
        each2timestep = self(f'cpu_per_each2{ct}')[timer]
        return guess_each * each2timestep

    @known_pattern(r'guess_cpu_seconds(|_unsafe)',
                   deps=['guess_cpu_seconds_sum_per_timestep_safe'])  # [TODO] deps based on group0
    def get_guess_cpu_seconds(self, var, *, _match=None):
        '''guess run cost, in cpu seconds. nt * guess_cpu_seconds_sum_per_timestep_safe
        Uses 'safe' guess (unless '_unsafe' included in name) but excludes initialization time.
            Might want to add a safety factor (e.g. assume 20% larger than this time).
        '''
        unsafe, = _match.groups()
        safe = '' if unsafe else '_safe'
        return self.input_deck['nt'] * self(f'guess_cpu_seconds_sum_per_timestep{safe}')

    @known_pattern(r'guess_runtime_seconds(|_unsafe)',
                   deps=['n_processors', {0: 'guess_cpu_seconds{group0}'}])
    def get_guess_runtime_seconds(self, var, *, _match=None):
        '''guess run cost, in wall clock time [seconds]. guess_cpu_seconds / n_processors.
        Uses 'safe' guess (unless '_unsafe' included in name) but excludes initialization time.
            Might want to add a safety factor (e.g. assume 20% larger than this time).

        Compare directly with self('steps_seconds').
        '''
        unsafe, = _match.groups()
        return self(f'guess_cpu_seconds{unsafe}') / self('n_processors')

    @known_pattern(r'guess_node_hours(|_unsafe)',
                   deps=['tasks_per_node', {0: 'guess_cpu_seconds{group0}'}])
    def get_guess_node_hours(self, var, *, _match=None):
        '''guess run cost, in node hours. guess_cpu_seconds * seconds2hours / tasks_per_node
        (seconds2hours = 1/3600)
        Uses 'safe' guess (unless '_unsafe' included in name) but excludes initialization time.
            Might want to add a safety factor (e.g. assume 20% larger than this time).
        '''
        unsafe, = _match.groups()
        seconds2hours = 1 / 3600
        return self(f'guess_cpu_seconds{unsafe}') * seconds2hours / self('tasks_per_node')



"""
File Purpose: misc. tools related to xarray
"""
import datetime as datetime
import inspect
from collections.abc import Callable

import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from ..arrays import np_dtype_object_to_str
from ..docs_tools import format_docstring
from ..os_tools import next_unique_name
from ..sentinels import UNSET
from ..timing import ProgressUpdater
from ...defaults import DEFAULTS
from ...errors import (
    InputError, InputMissingError, InputConflictError,
    DimensionKeyError,
)


### --------------------- Copying --------------------- ###

@pcAccessor.register('copy_kw')
def xarray_copy_kw(array, dims=None, *, name=None, array_to_dataset=False):
    '''return dict of info suitable for creating a similar array or dataset.
    result includes dims, coords, and attrs (unchanged, copied).

    dims: None, str, or iterable of str
        if provided, only include these dims (and related coords) in the result.
        Useful if only interested in some of the dims,
            e.g. if array has x,y,z,t dims but only want to mimic dims and coords from x,t.
    name: None, bool, or str
        whether to also include name in result
        None --> True if array has name, else False
        True --> name = array.name.
        str --> name = name.
    array_to_dataset: bool
        if True, result will be suitable for creating a Dataset based on the input DataArray.
        (Equivalent to using dims=[], name=False, and deleting 'dims' key from result.)
    '''
    # handle array_to_dataset mode:
    if array_to_dataset:
        result = xarray_copy_kw(array, dims=[], name=False, array_to_dataset=False)
        del result['dims']
        return result
    # 'usual' usage
    result = dict()
    if isinstance(dims, str): dims = [dims]
    if dims is None: dims = array.dims
    elif any(d not in array.dims for d in dims):
        errmsg = (f'cannot copy_kw for dims={dims!r} because some dims not in array.dims={array.dims}.')
        raise DimensionKeyError(errmsg)
    result['dims'] = dims
    coords = dict()
    for cname, cval in array.coords.items():
        if len(cval.dims) == 0:  # scalar coord; always keep!
            coords[cname] = cval
        elif len(cval.dims) == 1 and cval.dims[0] in dims:  # 1D coord, and relevant
            coords[cname] = cval
        elif all(d in dims for d in cval.dims):  # 2D+ coord, and all relevant dims exist
            coords[cname] = (tuple(cval.dims), cval.data)
    result['coords'] = coords
    attrs = array.attrs.copy()
    result['attrs'] = attrs
    if name is None:
        name = hasattr(array, 'name')
    if name == True:
        name = array.name
    if name != False:
        result['name'] = name
    return result

@pcAccessor.register('with_data', totype='array')
def xarray_with_data(xarray, data, dims=None, *, name=None):
    '''return data as an xr.DataArray with dims, coords, name, and attrs copied from xarray.
    Similar to DataArray.copy(data=data), but here allows a bit more control.

    dims: None, str, or iterable of str
        if provided, only include these dims (and related coords) in the result.
        Useful if only interested in some of the dims,
            e.g. if array has x,y,z,t dims but only want to mimic dims and coords from x,t.
    name: None, bool, or str
        whether to also include name in result
        None --> True if array has name, else False
        True --> name = array.name.
        str --> name = name.
    '''
    if not isinstance(xarray, xr.DataArray):
        raise InputError(f'xarray_with_data expects args[0] to be a DataArray, not {type(xarray)}.')
    kw = xarray_copy_kw(xarray, dims=dims, name=name)
    return xr.DataArray(data, **kw)


### --------------------- Converting Dataset to Array --------------------- ###

@pcAccessor.register('as_array')
def xarray_as_array(array):
    '''return array if DataArray, else array[var] if array is Dataset with one var.
    if array is Dataset with multiple vars, crash with InputError.
    '''
    if isinstance(array, xr.DataArray):
        return array
    elif isinstance(array, xr.Dataset):
        if len(array) == 1:
            return array[list(array)[0]]
        else:
            errmsg = ('as_array(Dataset) only possible if Dataset has exactly 1 data_var, '
                     f'but got Dataset with {len(array)} vars: {list(array.data_vars)}.')
            raise InputError(errmsg)
    else:
        raise InputError(f'cannot convert object of type {type(array)} to DataArray.')


### --------------------- Merging --------------------- ###

@format_docstring(print_freq=DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ,
                  ngroup=DEFAULTS.XARRAY_MERGE_NGROUP)
def xarray_merge(arrays, *, ngroup=UNSET, verbose=True, missing_name_ok=True, to_ds=None,
                 _updater=None, **kw_xr_merge):
    '''xr.merge() these xarray objects (probably DataArrays or Datasets),
    but maybe slightly more efficiently than just using xr.merge() directly.
    Also, has slightly more convenient default behaviors:
        - by default, allow name=None on inputs (treat all inputs with None name as same name)
        - by default, if all inputs are DataArray with same name, output DataArray not Dataset.

    ngroup: int>=2, None, or UNSET
        [EFF] max number of arrays per xr.merge() call.
        (Final results should be unaffected by ngroup, just affects internal strategy.)
        UNSET --> use DEFAULTS.XARRAY_MERGE_NGROUP (default: {ngroup}).
        None --> merge all arrays in a single xr.merge() call.
        int>=2 --> merge arrays in groups of size ngroup, then merge the merged groups.
            (This can be much faster than ngroup=None; see xarray_merge for details.)
    verbose: bool or int
        whether to print progress updates. (Ignored if N <= ngroup.)
        0: never print progress updates
        1: print updates every DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ seconds (default: {print_freq})
        2: print updates every file.
    missing_name_ok: bool
        whether to allow name=None as a valid name for input arrays.
        True --> treat all inputs with name=None as having the same name (None).
            (if result is a Dataset this can lead to a key in result being None.)
    to_ds: None
        whether result should be a Dataset.
        None --> DataArray if and only if all inputs are DataArrays with same name.
        True --> always return Dataset.
        False --> always return a DataArray (crash if any inputs are not DataArrays with same name).
    _updater: None or ProgressUpdater
        if provided, use this (instead of a new ProgressUpdater) to print updates, ignoring `verbose`.

    --- Timing notes: ---
        Tt seems like merging more than ~16 arrays at once can cause significant slowdowns,
        when compared to merging groups of 16, then merging those results.
        Tests with xarray 2024.7.0, Python 3.9.13, with subsets of 256 arrays totaling roughly 2 GB:
            64 arrays with N=None --> ~42 seconds | N=2 --> ~16s | N=4, 8, 16 --> ~12s
            128 arrays with N=None --> (didn't test it) | N=16 --> ~30 seconds
            256 arrays with N=None --> crashed due to memory errors | N=16 --> ~102 seconds
    '''
    # [TODO] is this basically just becoming xarray.combine_by_coords()?
    #   should that method be used instead? (Maybe it would be faster, too?)

    # check for any "obvious" crashes early:
    if len(arrays) == 0:
        raise InputMissingError('xarray_merge expects at least 1 DataArray or Dataset.')
    _to_ds_input = to_ds  # helps with debugging
    if not to_ds:
        _all_arrays = all(isinstance(arr, xr.DataArray) for arr in arrays)
        names = [arr.name for arr in arrays]
        _all_samename = len(set(names)) == 1
        if to_ds is None:
            to_ds = (not (_all_arrays and _all_samename))
        else:  # to_ds == False
            if not _all_arrays:
                raise InputConflictError('to_ds=False only possible if all inputs are DataArrays.')
            if not _all_samename:
                errmsg = f'to_ds=False requires all inputs to have the same name. Got: {names!r}.'
                raise InputConflictError(errmsg)
    if not missing_name_ok:
        if any(isinstance(arr, xr.DataArray) and arr.name is None for arr in arrays):
            raise InputConflictError('missing_name_ok=False but at least 1 input has name=None.')
    # handle simple case now (only 1 input)
    if len(arrays) == 1:
        result = arrays[0]
        if to_ds and isinstance(result, xr.DataArray):
            return xr.Dataset({result.name: result})
        else:
            return result
    # more bookkeeping
    names = set(arr.name for arr in arrays if isinstance(arr, xr.DataArray))
    rename_Nones = False
    if None in names:
        arrays = [*arrays]  # <-- copy input list to ensure input isn't edited in-place
        rename_Nones = next_unique_name('__xarray_merge_None_placeholder__', names)
        for i, arr in enumerate(arrays):
            if isinstance(arr, xr.DataArray) and arr.name is None:
                arrays[i] = arr.rename(rename_Nones)
    if ngroup is UNSET:
        ngroup = DEFAULTS.XARRAY_MERGE_NGROUP
    # >>> actual merging <<<
    if ngroup is None or len(arrays) <= ngroup:
        result = xr.merge(arrays, **kw_xr_merge)
    else:
        # else, need to do grouping
        print_freq = None if verbose==1 else (0 if verbose>=2 else -1)
        updater = ProgressUpdater(print_freq=print_freq) if _updater is None else _updater
        _orig_arrays = arrays  # just to help with debugging / printouts.
        layer = 0
        while len(arrays) > ngroup:
            merged = []
            for i0 in range(0, len(arrays), ngroup):
                i1 = min(i0 + ngroup, len(arrays))
                updater.print(f'Merging i={i0} to i={i1-1}, of max(i)={len(arrays)-1}' +
                              '.' if layer==0 else f' (-- merging merged arrays; at layer {layer}).')
                merged_group = xr.merge(arrays[i0:i1], **kw_xr_merge)
                merged.append(merged_group)
            arrays = merged
            layer = layer + 1
        # final merge:
        updater.print(f'Final merge of ({len(arrays)}) merged arrays.')
        result = xr.merge(arrays, **kw_xr_merge)
        updater.finalize(f'Merging of {len(_orig_arrays)} arrays complete.')
    # bookkeeping
    if rename_Nones:
        result = result.rename({rename_Nones: None})
    if not to_ds:   # (result of xr.merge() is always a Dataset. Convert back to DataArray.)
        key0 = next(iter(result.keys()))
        result = result[key0]
    return result

@pcAccessor.register('merge')
@format_docstring(xarray_merge_doc=xarray_merge.__doc__, sub_ntab=1)
def xarray_merge_from(array, other_arrays, *, ngroup=UNSET, verbose=True, **kw_xarray_merge):
    '''xr.merge() this xarray object (DataArray or Dataset) with other xarray objects.
    Internally, uses xarray_merge which might be a bit more efficient than xr.merge().

    xarray_merge docs copied below for convenience:
    -----------------------------------------------
        {xarray_merge_doc}
    '''
    all_arrays = [array, *other_arrays]
    return xarray_merge(all_arrays, ngroup=ngroup, **kw_xr_merge)


### --------------------- Vars from dataset --------------------- ###

@pcAccessor.register('vars_lookup', totype='dataset')
def xarray_vars_lookup(ds, lookup={}, **lookup_as_kw):
    '''return dict of {key from lookup: ds[lookup[key]]}
    result vals will be None for keys not found in ds.

    lookup: dict of {key: None, varname, or list of varnames}
        look for these varnames in ds. if list, use first varname found.
        if None, use key as varname.
    '''
    if lookup and lookup_as_kw:
        raise InputConflictError('provide lookup or lookup_as_kw, not both!')
    if lookup_as_kw:
        lookup = lookup_as_kw
    result = {}
    for key, find in lookup.items():
        result[key] = None
        if find is None:
            find = key
        if isinstance(find, str):
            find = [find]
        for varname in find:
            if varname in ds:
                result[key] = ds[varname]
                break
    return result

@pcAccessor.register('vars_lookup_with_defaults', totype='dataset')
def xarray_vars_lookup_with_defaults(ds, provided, lookup={}, defaults={}):
    '''return dict of {key from lookup: value for that key}.
    
    provided: dict of {key: None, str, or other value}
        provided values or lookup instructions for each key.
        None --> use instructions from lookup if key in lookup,
                 else value from defaults if key in defaults.
        str --> varname; result uses ds[varname]. (crash if not possible.)
        callable --> f of one arg; result uses f(ds).
        other value --> result uses this value directly.
    lookup: dict of {key: None, str, or list of str}
        'default' lookup instructions. Used when provided[key] is None.
    defaults: dict of {key: value or callable}
        default values for each key.
        Used when provided[key] is None and ds[lookup[key]] not found.
        callable --> called with ds as arg to get value.

    Examples:
        xarray_vars_lookup_with_defaults(ds, {}, {'n': None}, {'n': 100})
        # {'n': ds['n']} if ds['n'] exists, else {'n': 100}
        
        xarray_vars_lookup_with_defaults(ds, {}, {'n': 'var1'}, {'n': lambda ds: ds.size})
        # {'n': ds['var1']} if ds['var1'] exists, else {'n': ds.size}

        xarray_vars_lookup_with_defaults(ds, {'n': 'var1'}, {'n': 'var2'}, {'n': 7})
        # {'n': ds['var1']} if ds['var1'] exists, else crash.
        # (lookup & defaults are fully ignored, if non-None value in provided.)
    '''
    result = {}
    all_keys = set(provided) | set(lookup) | set(defaults)
    for key in all_keys:
        prov = provided.get(key, None)
        if prov is None:
            find = lookup.get(key, None)
            if find is None:
                find = key
            if isinstance(find, str):
                find = [find]
            for varname in find:
                if varname in ds:
                    result[key] = ds[varname]
                    break
            else:
                if key in defaults:
                    val = defaults[key]
                    if callable(val):
                        val = val(ds)
                    result[key] = val
                else:
                    errmsg = (f'no value for key={key!r}; not provided directly, '
                              f'varnames {find} not in ds, and no default provided.')
                    raise InputError(errmsg)
        elif isinstance(prov, str):
            if prov in ds:
                result[key] = ds[prov]
            else:
                errmsg = f'provided varname={prov!r} (for key={key!r}) not found in ds.'
                raise InputError(errmsg)
        elif callable(prov):
            result[key] = prov(ds)
        else:
            result[key] = prov
    return result


### --------------------- Math Checks --------------------- ###

@pcAccessor.register('where_finite')
def xarray_where_finite(array):
    '''returns array, masked with NaNs anywhere that the values are not finite.'''
    return array.where(np.isfinite)


### --------------------- Type Casting --------------------- ###

@pcAccessor.register('astypes', totype='dataset')
def xarray_astypes(ds, var_to_dtype):
    '''return copy of ds with var.astype(dtype) for each var, dtype in var_to_dtype.items()'''
    to_assign = {dvar: ds[dvar].astype(dtype) for dvar, dtype in var_to_dtype.items()}
    return ds.assign(**to_assign)

@pcAccessor.register('convert_types')
def xarray_convert_types(array, oldtype_to_newtype):
    '''return copy of array or dataset, using var.astype(newtype) for any var with oldtype,
    for oldtype, newtype in oldtype_to_newtype.items().
    '''
    if isinstance(array, xr.Dataset):
        ds = array
        to_convert = dict()
        for var, val in ds.items():
            for oldtype, newtype in oldtype_to_newtype.items():
                if val.dtype == oldtype:
                    to_convert[var] = newtype
                    break
        return xarray_astypes(ds, to_convert)
    else:
        for oldtype, newtype in oldtype_to_newtype.items():
            if array.dtype == oldtype:
                return array.astype(newtype)
        else:  # no match found during loop
            return array

@pcAccessor.register('dtype_object_to_str', totype='array')
def xarray_dtype_object_to_str(array):
    '''convert array with dtype=object to dtype=str, by doing str(x) for every x in array.'''
    return array.copy(data=np_dtype_object_to_str(array.values))


### --------------------- Converting coords to strings --------------------- ###

@pcAccessor.register('object_coords_to_str')
def xarray_object_coords_to_str(array, *, maxlen=None, ndim_min=1):
    '''return copy of array with coords (of dtype=object) converted to string.
    maxlen: None or int>=5
        if int, truncate longer strings to this length-3, and add '...' to end.
    ndim_min: int
        minimum number of dimensions for a coord to be converted.
        e.g. ndim_min=1 --> coords with ndim=0 will not be altered
    '''
    to_assign = {}
    for cname, cc in array.coords.items():
        if cc.dtype == object and cc.ndim >= ndim_min:
            if cc.ndim >= 2:
                raise NotImplementedError('object_coords_to_str does not yet support ndim>=2.')
            cvals = [str(v) for v in cc.values]
            if maxlen is not None:
                cvals = [v[:maxlen-3] + '...' if len(v) > maxlen else v for v in cvals]
            to_assign[cname] = cc.copy(data=cvals)
    return array.assign_coords(to_assign)



### --------------------- Recording "HISTORY" attr --------------------- ###

@pcAccessor.register('update_call_history', totype='dataset')
def xarray_update_call_history(ds: xr.Dataset, local_vars: dict, func: Callable) -> None:
    """
    Add a history entry to a dataset. Edits the input dataset in-place, then returns it.

    ds.attrs["HISTORY"] stores entries which tell a history of functions and their inputs,
        which were called and recorded in the dataset via update_call_history().
    The entries are strings, intended primarily for human-readability.

    Also adds ds.attrs["date created"] if it does not exist,
        else adds or updates ds.attrs["date modified"] to today's date.

    Parameters
    ----------
    ds : `xarray.Dataset`
        Dataset to update.
    local_vars : `dict`
        Local variables from the calling function.
    func : `Callable`
        Function or method being recorded in the history.
    """
    string_vals = []
    for arg, value in local_vars.items():
        if arg in inspect.signature(func).parameters:
            if isinstance(value, (xr.Dataset, xr.DataArray, np.ndarray)):
                if isinstance(value, np.ndarray) and value.shape in [(), (1,)]:
                    string_vals.append(f"{arg}={value.tolist()}")
                elif isinstance(value, xr.DataArray) and value.size == 1:
                    string_vals.append(f"{arg}={value.values.tolist()}")
                else:
                    string_vals.append(f"{arg}={arg}")
            else:
                string_vals.append(f"{arg}={value}")

    history_entry = f"{func.__name__}({', '.join(string_vals)})"
    if "HISTORY" in ds.attrs:
        if isinstance(ds.attrs["HISTORY"], list):
            ds.attrs["HISTORY"].append(history_entry)
        else:
            ds.attrs["HISTORY"] = [ds.attrs["HISTORY"], history_entry]
    else:
        ds.attrs["HISTORY"] = [history_entry]

    today = datetime.datetime.now(tz=datetime.timezone.utc)
    if "date created" in ds.attrs:
        ds.attrs["date modified"] = today.strftime("%d-%b-%Y")
    else:
        ds.attrs["date created"] = today.strftime("%d-%b-%Y")
    return ds



"""
File Purpose: tools related to xarray vector arithmetic along 'component' dimension,
probably with 'x', 'y', 'z' components.

NOTE: functions with multiple vector inputs here are not registered here with pcAccessor.
    See vector_arithmetic.py for similar versions which ARE registered with pcAccessor.
    (Those versions might eventually be updated to reference-frame-aware checks,
    taking Component machinery under consideration.
    The versions here will instead never do any implicit reference frame rotations.)
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_dimensions import (
    take_along_dimension, join_along_dimension,
    xarray_assign,
)
from .xarray_indexing import xarray_sel
from ..sentinels import UNSET
from ...errors import InputError

### --------------------- Allow vector tools without Component --------------------- ###
# Tools should all work properly whether _BASIS is still strings, or proper Component objects.
# It will be strings if ported these tools via copy-pasting.
# It will be Component objects if used inside PlasmaCalcs, as components.py updates _BASIS.

class BasisXYZ():
    '''simple class to hold X, Y, Z component details.
    (Assumed to be immutable after creation)

    XYZ: list-like with length 3
        iterable containing [X, Y, Z] component labels.

    Instances of this class provide attributes:
        X: x component label
        Y: y component label
        Z: z component label
        XHAT: x unit vector, as a DataArray of [1,0,0] with 'component' values XYZ
        YHAT: y unit vector, as a DataArray of [0,1,0] with 'component' values XYZ
        ZHAT: z unit vector, as a DataArray of [0,0,1] with 'component' values XYZ
        YZ_FROM_X: dict telling (y,z) given x, for right-handed coord system x,y,z
            (internally here: {X:(Y,Z), Y:(Z,X), Z:(X,Y)})
    '''
    def __init__(self, XYZ):
        self.XYZ = XYZ
        self.X = self.XYZ[0]
        self.Y = self.XYZ[1]
        self.Z = self.XYZ[2]
        self.XHAT = xr.DataArray([1,0,0], coords={'component': self.XYZ})
        self.YHAT = xr.DataArray([0,1,0], coords={'component': self.XYZ})
        self.ZHAT = xr.DataArray([0,0,1], coords={'component': self.XYZ})
        self.YZ_FROM_X = {
            self.X: (self.Y, self.Z),
            self.Y: (self.Z, self.X),
            self.Z: (self.X, self.Y),
        }

    def get(self, x):
        '''return proper component label (i.e. from self.XYZ) given x.
        This is the first match from XYZ which equals x. Crash if no match.
        '''
        for xc in self.XYZ:
            if xc == x:
                return xc

    def get_hat(self, x):
        '''return self.XHAT, self.YHAT, or self.ZHAT depending on str(self.get(x)).'''
        _x_input = x
        x = str(self.get(x))
        if x == str(self.X):
            return self.XHAT
        elif x == str(self.Y):
            return self.YHAT
        elif x == str(self.Z):
            return self.ZHAT
        else:
            errmsg = (f'{type(self).__name__}.hat(x={_x_input!r}), but str(self.get(x))=={x!r} not found in: '
                      f'str(self.X)={str(self.X)!r}, str(self.Y)={str(self.Y)!r}, str(self.Z)={str(self.Z)!r}.')
            raise InputError(errmsg)

    def __repr__(self):
        return f'{type(self).__name__}(XYZ={self.XYZ})'

_BASIS = BasisXYZ(['x', 'y', 'z'])  # default basis, using strings
# CAUTION: don't rely on _BASIS for kwarg defaults, only use it inside functions.

def _get_basis(_basis=UNSET):
    '''return proper BasisXYZ to use.
    _basis: UNSET or BasisXYZ
        tells basis to use.
        UNSET --> default to xarray_vectors._BASIS
    '''
    return _BASIS if _basis is UNSET else _basis


### --------------------- Independent of _BASIS --------------------- ###

# @pcAccessor.register('dot_product', aliases=['dot'])  # registered in vector_arithmetic.py instead
def xarray_dot_product(A, B):
    '''return dot product of vectors A and B, assuming vector components along the dimension 'component'.'''
    # note: need to ensure to only sum across matching components
    #    (can't just take_along_dimension('component') without checking if components match)
    # also note: (A * B).sum('component') fails when component is a non-dimension coordinate,
    #    e.g. when there is only 1 component.
    # [TODO] probably a more concise way to write this function, using xarray_promote_dim.
    A_dim = 'component' in A.dims
    B_dim = 'component' in B.dims
    # case: both have 'component' dimension. Easy - let xarray handle all the bookkeeping.
    if A_dim and B_dim:
        result = A * B
        return result.sum('component', skipna=False)  # skipna=False --> result NaN wherever any summand is NaN
    # case: exactly one has 'component' dimension but the other does not.
    elif A_dim or B_dim:
        if A_dim:
            a, b = A, B
        else:
            a, b = B, A
        # now, 'a' has 'component' dimension while 'b' does not.
        if b.coords['component'].size != 1:
            raise NotImplementedError('dot_product, when component is a non-dimension coordinate with size > 1')
        idx = np.nonzero(a.coords['component'].values == b.coords['component'].values)[0]
        if len(idx) == 0:  # no matching components. Result is 0.
            return xr.zeros_like(b).drop_vars('component')  # drop_vars --> result has no 'component' coord.
        else:
            assert len(idx) == 1
            i = idx[0]
            result = a.isel(component=i) * b
            return result.drop_vars('component')  # drop_vars --> result has no 'component' coord.
    # case: A and B both do not have 'component' dimension.
    else:
        # first ensure A and B have the same 'component' coordinate, then it's easy, just multiply them.
        if not (A.coords['component'].size == 1 == B.coords['component'].size):
            raise NotImplementedError('dot_product, when component is a non-dimension coordinate with size > 1')
        A_comps = A.coords['component'].values
        B_comps = B.coords['component'].values
        if np.all(A_comps == B_comps):
            result = A * B
            return result.drop_vars('component')  # drop_vars --> result has no 'component' coord.
        else:
            return xr.zeros_like(A).drop_vars('component')  # drop_vars --> result has no 'component' coord.
    assert False, "if reached this line, made coding mistake."

@pcAccessor.register('magnitude', aliases=['mag', 'mod'])
def xarray_magnitude(A, *, squared=False):
    '''return vector magnitude of A, assuming vector components along the dimension 'component'.
    squared: bool, default False
        if True, return |A|**2 instead of |A|.
        [EFF] to get |A|**2, when |A| is not needed,
            magnitude(A, squared=True) is more efficient than magnitude(A)**2
    '''
    result = xarray_dot_product(A, A)
    return result if squared else result**0.5

@pcAccessor.register('unit_vector', aliases=['hat', 'dir', 'direction'])
def xarray_unit_vector(A):
    '''return a unit vector in the direction of A. Equivalent: A / |A|'''
    return A / xarray_magnitude(A)

@pcAccessor.register('rmscomps')
def xarray_rmscomps(A):
    '''return root mean squared of components of A.
    E.g., rmscomps(A) --> sqrt((Ax^2 + Ay^2 + Az^2) / 3), if A has 3 components.
    '''
    num_comps = A.coords['component'].size  # number of components
    mag2 = xarray_magnitude(A, squared=True)
    return np.sqrt(mag2 / num_comps)

# intentionally no @pcAccessor; see xarray_perp below instead.
def xarray_take_perp_to(B, A):
    '''return the component of A perpendicular to B. Equivalent: A - (A dot Bhat) Bhat.
    Note that B is the first argument.
    '''
    Bhat = B / xarray_magnitude(B)
    return A - xarray_dot_product(A, Bhat) * Bhat

# intentionally no @pcAccessor; see xarray_parallel below instead.
def xarray_take_parallel_to(B, A):
    '''return the component of A parallel to B. Equivalent: (A dot Bhat) Bhat.
    Note that B is the first argument.
    '''
    Bhat = B / xarray_magnitude(B)
    return xarray_dot_product(A, Bhat) * Bhat

# @pcAccessor.register('perp')  # in vector_arithmetic.py instead
def xarray_perp(A, B):
    '''return the component of A perpendicular to B. Equivalent: A - (A dot Bhat) Bhat.'''
    return xarray_take_perp_to(B, A)

# @pcAccessor.register('parallel', aliases=['par'])  # in vector_arithmetic.py instead
def xarray_parallel(A, B):
    '''return the component of A parallel to B. Equivalent: (A dot Bhat) Bhat.'''
    return xarray_take_parallel_to(B, A)

# @pcAccessor.register('perpmod', aliases=['perpmag'])  # in vector_arithmetic.py instead
def xarray_perpmod(A, B):
    '''return the magnitude of the component of A perpendicular to B. Equivalent: |A - (A dot Bhat) Bhat|'''
    return xarray_magnitude(perp(A, B))

# @pcAccessor.register('parmod', aliases=['parmag'])  # in vector_arithmetic.py instead
def xarray_parmod(A, B, *, signed=False):
    '''return the magnitude of the component of A parallel to B. Equivalent: |(A dot Bhat) Bhat|
    signed: bool; whether to return signed result instead, i.e. (A dot Bhat), without taking abs.
    '''
    result = xarray_dot_product(A, xarray_unit_vector(B))
    return result if signed else np.abs(result)



### --------------------- Related to / Depends on _BASIS --------------------- ###


@pcAccessor.register('take_xyz')
def xarray_take_xyz(A, *, drop=False, as_dict=False):  
    '''return [Ax, Ay, Az] from selecting 'x', 'y', 'z' along 'component' dimension.
    Crashes if any components are missing.

    drop: bool
        whether to drop x, y, z 'component' coord from results.
    as_dict: bool
        if True, return dict(x=Ax, y=Ay, z=Az) instead of list [Ax, Ay, Az].
    '''
    x = xarray_sel(A, component='x', drop=drop)
    y = xarray_sel(A, component='y', drop=drop)
    z = xarray_sel(A, component='z', drop=drop)
    return dict(x=x, y=y, z=z) if as_dict else [x, y, z]

@pcAccessor.register('join_xyz', aliases=['join_yz'])
def xarray_join_xyz(x, y, z, *, _basis=UNSET):
    '''return vector with components x, y, z along 'component' dimension.
    (Call as xarray_join_xyz(x, y, z), or x.pc.join_yz(y,z), or x.pc.join_xyz(y,z))
    Result has 'component' coord with values 'x', 'y', 'z'
        (or Component objects representing 'x', 'y', 'z', depending on _basis.)

    x, y, z: xarray.DataArray or Dataset objects
        vector components to use.
    _basis: UNSET or BasisXYZ
        tells basis from which to infer X, Y, Z component labels.
        UNSET --> default to xarray_vectors._BASIS
    '''
    if _basis is UNSET: _basis = _BASIS
    x = x.assign_coords(component=_basis.X)
    y = y.assign_coords(component=_basis.Y)
    z = z.assign_coords(component=_basis.Z)
    return join_along_dimension('component', [x, y, z])

# @pcAccessor.register('angle_xy', aliases=...)  # in vector_arithmetic.py instead
def xarray_angle_xy(A):   # okay this one doesn't actually depend on _BASIS but it comes pretty close...
    '''return angle between +xhat and A, in the xy plane, in radians.
    A should be a DataArray (or Dataset) with x and y in 'component' dimension.
    A can be any vector (does not need to be a unit vector, but it can be.)
    '''
    x = xarray_sel(A, component='x', drop=True)
    y = xarray_sel(A, component='y', drop=True)
    return np.arctan2(y, x)

# @pcAccessor.register('angle_xy_to_hat', ...)  # in vector_arithmetic.py instead
def xarray_angle_xy_to_hat(A, *, _basis=UNSET):
    '''return unit vector u, given angle [radians] between +xhat and u in the xy plane.
    The result has 3 components (x,y,z) with value equal to 0 for z component.
    Equivalent: cos(A) * xhat + sin(A) * yhat.

    _basis: UNSET or BasisXYZ
        tells basis from which to infer XHAT and YHAT
        UNSET --> default to xarray_vectors._BASIS
    '''
    if _basis is UNSET: _basis = _BASIS
    return np.cos(A) * _basis.XHAT + np.sin(A) * _basis.YHAT

# @pcAccessor.register('cross_component')  # in vector_arithmetic.py instead
def xarray_cross_component(A, B, x, *, yz=None, missing_ok=False, _basis=UNSET):
    '''return x component of A cross B, given A and B which have values for y and z 'component'.

    x: str (or any other valid component specifier)
        probably 'x', 'y', or 'z'; tells which component of A cross B to return.
        will be "normalized" via _basis.get(x).
    A, B: xarray.DataArray
        vectors to take cross product of.
        must include 'component' dimension including coordinates y and z.
    yz: None or iterable of two strings objects
        the other two components; (x,y,z) should form a right-handed coordinate system.
        if not provided, infer from x.
    missing_ok: bool, default False
        whether it is okay for 'component' dimension to be missing y or z components, of A or B.
        if True, treat any missing components as 0.
    _basis: UNSET or BasisXYZ
        tells basis from which to normalize x and to infer YZ_FROM_X mapping.
        UNSET --> default to xarray_vectors._BASIS
    '''
    if _basis is UNSET: _basis = _BASIS
    x = _basis.get(x)
    if yz is None: yz = _basis.YZ_FROM_X[x]
    default = 0 if missing_ok else UNSET
    Ay, Az = take_along_dimension('component', A, yz, default=default)
    By, Bz = take_along_dimension('component', B, yz, default=default)
    return xarray_assign(Ay * Bz - Az * By, coords={'component': x})

# @pcAccessor.register('cross_product')  # in vector_arithmetic.py instead
def xarray_cross_product(A, B, *, components=None, _basis=UNSET):
    '''return cross product of vectors A and B, along dimension 'component'.
    If A or B missing any components, treat them as 0.

    components: None or iterable of str (or any other valid component specifier)
        tells which components to get.
        None --> get all components (XYZ)
        will be "normalized" via _basis.get(x) for each component (see xarray_cross_component)
    _basis: UNSET or BasisXYZ
        tells basis from which to infer XYZ component labels.
        UNSET --> default to xarray_vectors._BASIS
    '''
    if _basis is UNSET: _basis = _BASIS
    if components is None:
        components = _basis.XYZ
    result = [xarray_cross_component(A, B, x, missing_ok=True) for x in components]
    return join_along_dimension('component', result)
